/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "CodeGeneratorPSP.h"
#include "N64RegisterCachePSP.h"
#include "AssemblyUtils.h"
#include "Trace.h"

#include "Core/R4300.h"
#include "Core/R4300_Regs.h"
#include "Core/CPU.h"
#include "Core/Memory.h"

#include "Debug/DBGConsole.h"

#include <algorithm>

using namespace AssemblyUtils;

//#define NAIVE_REGISTER_ALLOCATION

#define NOT_IMPLEMENTED( x )	DAEDALUS_ERROR( x )

extern "C" { void _CallReturnFromDynaRec(); } 
extern "C" { void _DirectExitCheck( u32 instructions_executed, u32 target_pc, u32 delay ); } 
extern "C" { void _IndirectExitCheck( u32 instructions_executed, CIndirectExitMap * p_map, u32 target_pc ); } 

namespace
{
const EPspReg	gRegistersToUseForCaching[] = 
{
	PspReg_S0,
	PspReg_S1,
	PspReg_S2,
	PspReg_S3,
	PspReg_S4,
	PspReg_S5,
	PspReg_S6,
//	PspReg_S7,		// Used for read address table
	PspReg_T2,
	PspReg_T3,
	PspReg_T4,
	PspReg_T5,
	PspReg_T6,
	PspReg_T7,
	PspReg_T8,
	PspReg_T9,
	PspReg_A2,
	PspReg_A3,
};
}

u32		gTotalRegistersCached = 0;
u32		gTotalRegistersUncached = 0;


//*****************************************************************************
//
//*****************************************************************************
CCodeGeneratorPSP::CCodeGeneratorPSP( CAssemblyBuffer * p_buffer )
:	CCodeGenerator( )
,	CAssemblyWriterPSP( p_buffer )
,	mpBasePointer( NULL )
,	mBaseRegister( PspReg_S8 )		// TODO
{
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::Finalise()
{
	SetAssemblyBuffer( NULL );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::AnalyseStats( const SRegisterStats & stats )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetCachedBasePointer( const void * p_base )
{
	mpBasePointer = reinterpret_cast< const u8 * >( p_base );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetRegisterSpanList( const RegisterSpanList & span_list )
{

	mRegisterSpanList = span_list;

	// Sort in order of increasing start point
	std::sort( mRegisterSpanList.begin(), mRegisterSpanList.end(), SAscendingSpanStartSort() );

	const u32 NUM_CACHE_REGS( sizeof(gRegistersToUseForCaching) / sizeof(gRegistersToUseForCaching[0]) );

	// Push all the available registers in reverse order (i.e. use temporaries later)
	DAEDALUS_ASSERT( mAvailableRegisters.empty(), "Why isn't the available register list empty?" );
	for( s32 i = NUM_CACHE_REGS-1; i >= 0 ; --i )
	{
		mAvailableRegisters.push( gRegistersToUseForCaching[ i ] );
	}

#ifdef NAIVE_REGISTER_ALLOCATION
	u32		cache_reg_idx( 0 );
	for(RegisterSpanList::const_iterator span_it = mRegisterSpanList.begin(); span_it < mRegisterSpanList.end(); ++span_it )
	{
		const SRegisterSpan &	span( *span_it );

		if( cache_reg_idx < NUM_CACHE_REGS )
		{
			EPspReg		cachable_reg( gRegistersToUseForCaching[cache_reg_idx] );

			mRegisterCache.SetCachedReg( span.Register, 0, cachable_reg );
			cache_reg_idx++;
		}

	}

	// If there are still unused registers, assign to high part of reg
	if( cache_reg_idx < NUM_CACHE_REGS )
	{
		for(RegisterSpanList::const_iterator span_it = mRegisterSpanList.begin(); span_it < mRegisterSpanList.end(); ++span_it )
		{
			const SRegisterSpan &	span( *span_it );

			if( cache_reg_idx < NUM_CACHE_REGS )
			{
				EPspReg		cachable_reg( gRegistersToUseForCaching[cache_reg_idx] );

				mRegisterCache.SetCachedReg( span.Register, 1, cachable_reg );
				cache_reg_idx++;
			}
		}
	}
#endif
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::ExpireOldIntervals( u32 instruction_idx )
{
	// mActiveIntervals is held in order of increasing end point
	for(RegisterSpanList::iterator span_it = mActiveIntervals.begin(); span_it < mActiveIntervals.end(); ++span_it )
	{
		const SRegisterSpan &	span( *span_it );

		if( span.SpanEnd >= instruction_idx )
		{
			break;
		}

		// This interval is no longer active - flush the register and return it to the list of available regs
		EPspReg		psp_reg( mRegisterCache.GetCachedReg( span.Register, 0 ) );

		FlushAndInvalidateRegister( span.Register, 0 );

		mRegisterCache.ClearCachedReg( span.Register, 0 );

		mAvailableRegisters.push( psp_reg );

		span_it = mActiveIntervals.erase( span_it );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SpillAtInterval( const SRegisterSpan & live_span )
{
	DAEDALUS_ASSERT( !mActiveIntervals.empty(), "There are no active intervals" );

	const SRegisterSpan &	last_span( mActiveIntervals.back() );		// Spill the last active interval (it has the greatest end point)

	if( last_span.SpanEnd > live_span.SpanEnd )
	{
		// Uncache the old span
		EPspReg		psp_reg( mRegisterCache.GetCachedReg( last_span.Register, 0 ) );
		FlushAndInvalidateRegister( last_span.Register, 0 );
		mRegisterCache.ClearCachedReg( last_span.Register, 0 );

		// Cache the new span
		mRegisterCache.SetCachedReg( live_span.Register, 0, psp_reg );

		mActiveIntervals.pop_back();				// Remove the last span
		mActiveIntervals.push_back( live_span );	// Insert in order of increasing end point

		std::sort( mActiveIntervals.begin(), mActiveIntervals.end(), SAscendingSpanEndSort() );		// XXXX - will be quicker to insert in the correct place rather than sorting each time
	}
	else
	{
		// There is no space for this register - we just don't update the register cache info, so we save/restore it from memory as needed
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::UpdateRegisterCaching( u32 instruction_idx )
{
#ifndef NAIVE_REGISTER_ALLOCATION
	ExpireOldIntervals( instruction_idx );

	for(RegisterSpanList::const_iterator span_it = mRegisterSpanList.begin(); span_it < mRegisterSpanList.end(); ++span_it )
	{
		const SRegisterSpan &	span( *span_it );

		// As we keep the intervals sorted in order of SpanStart, we can exit as soon as we encounter a SpanStart in the future
		if( instruction_idx < span.SpanStart )
		{
			break;
		}

		// Only process live intervals
		if( instruction_idx >= span.SpanStart && instruction_idx <= span.SpanEnd )
		{
			if( !mRegisterCache.IsCached( span.Register, 0 ) )
			{
				if( mAvailableRegisters.empty() )
				{
					SpillAtInterval( span );
				}
				else
				{
					// Use this register for caching
					mRegisterCache.SetCachedReg( span.Register, 0, mAvailableRegisters.top() );

					// Pop this register from the available list
					mAvailableRegisters.pop();
					mActiveIntervals.push_back( span );		// Insert in order of increasing end point

					std::sort( mActiveIntervals.begin(), mActiveIntervals.end(), SAscendingSpanEndSort() );		// XXXX - will be quicker to insert in the correct place rather than sorting each time
				}
			}
		}
	}
#endif
}

//*****************************************************************************
//
//*****************************************************************************
RegisterSnapshotHandle	CCodeGeneratorPSP::GetRegisterSnapshot()
{
	RegisterSnapshotHandle	handle( mRegisterSnapshots.size() );
	
	mRegisterSnapshots.push_back( mRegisterCache );

	return handle;
}

//*****************************************************************************
//
//*****************************************************************************
CCodeLabel	CCodeGeneratorPSP::GetEntryPoint() const
{
	return GetAssemblyBuffer()->GetStartAddress();
}

//*****************************************************************************
//
//*****************************************************************************
u32	CCodeGeneratorPSP::GetCompiledCodeSize() const
{
	return GetAssemblyBuffer()->GetSize();
}

//*****************************************************************************
//
//*****************************************************************************
EPspReg	CCodeGeneratorPSP::GetRegisterNoLoad( EN64Reg n64_reg, u32 lo_hi_idx, EPspReg scratch_reg )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		return mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx );
	}
	else
	{
		return scratch_reg;
	}
}

//*****************************************************************************
//
//*****************************************************************************
EPspReg	CCodeGeneratorPSP::GetRegisterAndLoad( EN64Reg n64_reg, u32 lo_hi_idx, EPspReg scratch_reg )
{
	EPspReg		reg;
	bool		need_load( false );

	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		gTotalRegistersCached++;
		reg = mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx );

		// We're loading it below, so set the valid flag
		if( !mRegisterCache.IsValid( n64_reg, lo_hi_idx ) )
		{
			need_load = true;
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
			DAEDALUS_ASSERT( !mRegisterCache.IsDirty( n64_reg, lo_hi_idx ), "Shouldn't be set as dirty if just loaded" );
		}
	}
	else if( n64_reg == N64Reg_R0 )
	{
		reg = PspReg_R0;
	}
	else
	{
		gTotalRegistersUncached++;
		reg = scratch_reg;
		need_load = true;
	}

	if( need_load )
	{
		GetVar( reg, &gGPR[ n64_reg ]._u32[ lo_hi_idx ] );
	}

	return reg;
}

//*****************************************************************************
//	Similar to GetRegisterAndLoad, but ALWAYS loads into the specified psp register
//*****************************************************************************
void CCodeGeneratorPSP::LoadRegister( EPspReg psp_reg, EN64Reg n64_reg, u32 lo_hi_idx )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		gTotalRegistersCached++;

		// Load the register if it's currently invalid
		if( !mRegisterCache.IsValid( n64_reg,lo_hi_idx ) )
		{
			GetVar( cached_reg, &gGPR[ n64_reg ]._u32[ lo_hi_idx ] );
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
			DAEDALUS_ASSERT( !mRegisterCache.IsDirty( n64_reg, lo_hi_idx ), "Shouldn't be set as dirty if just loaded" );
		}

		// Copy the register if necessary
		if( psp_reg != cached_reg )
		{
			OR( psp_reg, cached_reg, PspReg_R0 );
		}
	}
	else if( n64_reg == N64Reg_R0 )
	{
		OR( psp_reg, PspReg_R0, PspReg_R0 );
	}
	else
	{
		gTotalRegistersUncached++;
		GetVar( psp_reg, &gGPR[ n64_reg ]._u32[ lo_hi_idx ] );
	}
}

//*****************************************************************************
//	This function pulls in a cached register so that it can be used at a later point.
//	This is usally done when we have a branching instruction - it guarantees that
//	the register is valid regardless of whether or not the branch is taken.
//*****************************************************************************
void	CCodeGeneratorPSP::PrepareCachedRegister( EN64Reg n64_reg, u32 lo_hi_idx )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		// Load the register if it's currently invalid
		if( !mRegisterCache.IsValid( n64_reg,lo_hi_idx ) )
		{
			GetVar( cached_reg, &gGPR[ n64_reg ]._u32[ lo_hi_idx ] );
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
			DAEDALUS_ASSERT( !mRegisterCache.IsDirty( n64_reg, lo_hi_idx ), "Shouldn't be set as dirty if just loaded" );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::StoreRegister( EN64Reg n64_reg, u32 lo_hi_idx, EPspReg psp_reg )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		gTotalRegistersCached++;

		// Update our copy as necessary
		if( psp_reg != cached_reg )
		{
			OR( cached_reg, psp_reg, PspReg_R0 );
		}
		mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, true );
		mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
	}
	else
	{
		gTotalRegistersUncached++;

		SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], psp_reg );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::SetRegister( EN64Reg n64_reg, u32 lo_hi_idx, u32 value, EPspReg scratch_reg )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		gTotalRegistersCached++;

		LoadConstant( cached_reg, value );
		mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, true );
		mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
	}
	else
	{
		EPspReg		reg;
		if( value == 0 )
		{
			reg = PspReg_R0;
		}
		else
		{
			LoadConstant( scratch_reg, value );
			reg = scratch_reg;
		}
		gTotalRegistersUncached++;

		SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], reg );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::UpdateRegister( EN64Reg n64_reg, EPspReg psp_reg, EUpdateRegOptions options, EPspReg scratch_reg )
{
	StoreRegisterLo( n64_reg, psp_reg );

	switch( options )
	{
	case URO_HI_LEAVE:
		break;
	case URO_HI_SIGN_EXTEND:
		SRA( scratch_reg, psp_reg, 0x1f );		// Sign extend
		StoreRegisterHi( n64_reg, scratch_reg );
		break;
	case URO_HI_CLEAR:
		SetRegisterHi( n64_reg, 0, scratch_reg );
		break;
	case URO_HI_SET:
		SetRegisterHi( n64_reg, 0xffffffff, scratch_reg );
		break;
	default:
		DAEDALUS_ERROR( "Unhandled option" );
		NODEFAULT;
		break;
	}
}

//*****************************************************************************
//
//*****************************************************************************
EPspFloatReg	CCodeGeneratorPSP::GetFloatRegisterNoLoad( EN64FloatReg n64_reg )
{
	return EPspFloatReg( n64_reg );	// 1:1 mapping
}

//*****************************************************************************
//
//*****************************************************************************
EPspFloatReg	CCodeGeneratorPSP::GetFloatRegisterAndLoad( EN64FloatReg n64_reg )
{
	EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );	// 1:1 mapping
	if( !mRegisterCache.IsFPValid( n64_reg ) )
	{
		GetFloatVar( psp_reg, &g_qwCPR[1][n64_reg]._f32[ 0 ] );
		mRegisterCache.MarkFPAsValid( n64_reg, true );
	}

	return psp_reg;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::UpdateFloatRegister( EN64FloatReg n64_reg )
{
	mRegisterCache.MarkFPAsDirty( n64_reg, true );
	mRegisterCache.MarkFPAsValid( n64_reg, true );
}

//*****************************************************************************
//
//*****************************************************************************
const CN64RegisterCachePSP & CCodeGeneratorPSP::GetRegisterCacheFromHandle( RegisterSnapshotHandle snapshot ) const
{
	DAEDALUS_ASSERT( snapshot.Handle < mRegisterSnapshots.size(), "Invalid snapshot handle" );
	return mRegisterSnapshots[ snapshot.Handle ];
}

//*****************************************************************************
//	This function flushes all dirty registers back to memory, but leaves the valid flag untouched
//*****************************************************************************
void	CCodeGeneratorPSP::FlushAllRegisters()
{
	for( u32 lo_hi_idx = 0; lo_hi_idx < 2; ++lo_hi_idx )
	{
		// Skip r0
		for( u32 i = 1; i < NUM_N64_REGS; ++i )
		{
			EN64Reg	n64_reg = EN64Reg( i );
			if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
			{
				if( mRegisterCache.IsDirty( n64_reg, lo_hi_idx ) )
				{
					DAEDALUS_ASSERT( mRegisterCache.IsValid( n64_reg, lo_hi_idx ), "Register is dirty but not valid?" );

					EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

					SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], cached_reg );

					mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, false );
				}
			}
		}
	}

	for( u32 i = 0; i < NUM_N64_FP_REGS; ++i )
	{
		EN64FloatReg	n64_reg = EN64FloatReg( i );
		if( mRegisterCache.IsFPDirty( n64_reg ) )
		{
			DAEDALUS_ASSERT( mRegisterCache.IsFPValid( n64_reg ), "Register is dirty but not valid?" );

			EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );

			SetFloatVar( &g_qwCPR[1][n64_reg]._f32[ 0 ], psp_reg );

			mRegisterCache.MarkFPAsDirty( n64_reg, false );
		}
	}
}

//*****************************************************************************
//	This function flushes all dirty registers back to memory, and marks them
//	as invalid. This is primarily to ensure that we keep the register set
//	in a consistent set across calls to generic functions. Ideally we need
//	to reimplement generic functions with specialised code to avoid the flush.
//*****************************************************************************
void	CCodeGeneratorPSP::FlushAndInvalidateAllRegisters()
{
	for( u32 lo_hi_idx = 0; lo_hi_idx < 2; ++lo_hi_idx )
	{
		// Skip r0
		for( u32 i = 1; i < NUM_N64_REGS; ++i )
		{
			EN64Reg	n64_reg = EN64Reg( i );
			if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
			{
				if( mRegisterCache.IsDirty( n64_reg, lo_hi_idx ) )
				{
					DAEDALUS_ASSERT( mRegisterCache.IsValid( n64_reg, lo_hi_idx ), "Register is dirty but not valid?" );

					EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

					SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], cached_reg );

					mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, false );
				}

				// Invalidate the register, so we pick up any values the function might have changed
				mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, false );
			}
		}
	}

	for( u32 i = 0; i < NUM_N64_FP_REGS; ++i )
	{
		EN64FloatReg	n64_reg = EN64FloatReg( i );
		if( mRegisterCache.IsFPDirty( n64_reg ) )
		{
			DAEDALUS_ASSERT( mRegisterCache.IsFPValid( n64_reg ), "Register is dirty but not valid?" );

			EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );

			SetFloatVar( &g_qwCPR[1][n64_reg]._f32[ 0 ], psp_reg );

			mRegisterCache.MarkFPAsDirty( n64_reg, false );
		}

		// Invalidate the register, so we pick up any values the function might have changed
		mRegisterCache.MarkFPAsValid( n64_reg, false );
	}
}

//*****************************************************************************
//	This function flushes all dirty *temporary* registers back to memory, and
//	marks them as invalid.
//*****************************************************************************
void	CCodeGeneratorPSP::FlushAndInvalidateAllTemporaryRegisters()
{
	for( u32 lo_hi_idx = 0; lo_hi_idx < 2; ++lo_hi_idx )
	{
		// Skip r0
		for( u32 i = 1; i < NUM_N64_REGS; ++i )
		{
			EN64Reg	n64_reg = EN64Reg( i );
			if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
			{
				if( mRegisterCache.IsTemporary( n64_reg, lo_hi_idx ) )
				{
					if( mRegisterCache.IsDirty( n64_reg, lo_hi_idx ) )
					{
						DAEDALUS_ASSERT( mRegisterCache.IsValid( n64_reg, lo_hi_idx ), "Register is dirty but not valid?" );

						EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

						SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], cached_reg );

						mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, false );
					}

					// Invalidate the register, so we pick up any values the function might have changed
					mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, false );
				}
			}
		}
	}

	// XXXX some fp regs are preserved across function calls?
	for( u32 i = 0; i < NUM_N64_FP_REGS; ++i )
	{
		EN64FloatReg	n64_reg = EN64FloatReg( i );
		if( mRegisterCache.IsFPDirty( n64_reg ) )
		{
			DAEDALUS_ASSERT( mRegisterCache.IsFPValid( n64_reg ), "Register is dirty but not valid?" );

			EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );

			SetFloatVar( &g_qwCPR[1][n64_reg]._f32[ 0 ], psp_reg );

			mRegisterCache.MarkFPAsDirty( n64_reg, false );
		}

		// Invalidate the register, so we pick up any values the function might have changed
		mRegisterCache.MarkFPAsValid( n64_reg, false );
	}
}

//*****************************************************************************
//	Write back the specified register (if dirty) and mark as invalid
//*****************************************************************************
void	CCodeGeneratorPSP::FlushAndInvalidateRegister( EN64Reg n64_reg, u32 lo_hi_idx )
{
	DAEDALUS_ASSERT( mRegisterCache.IsCached( n64_reg, lo_hi_idx ), "This register is not cached" );

	if( mRegisterCache.IsDirty( n64_reg, lo_hi_idx ) )
	{
		DAEDALUS_ASSERT( mRegisterCache.IsValid( n64_reg, lo_hi_idx ), "Register is dirty but not valid?" );

		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], cached_reg );

		mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, false );
	}

	// Invalidate the register, so we pick up any values the function might have changed
	mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, false );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::GeneratePrologue()
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
// EPILOGUE - set up g_pPCMemBase before returning
//*****************************************************************************
CCodeLabel CCodeGeneratorPSP::GenerateEpilogue( u32 pc )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
	return CCodeLabel(NULL);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::CallUpdateCounter( u32 num_instructions )
{
	PspOpCode	op1;
	PspOpCode	op2;

	GetLoadConstantOps( PspReg_A0, num_instructions, &op1, &op2 );

	if( op2._u32 == 0 )
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( CPU_UpdateCounter ) ), false );		// No branch delay
		AppendOp( op1 );
	}
	else
	{
		AppendOp( op1 );
		JAL( CCodeLabel( reinterpret_cast< const void * >( CPU_UpdateCounter ) ), false );		// No branch delay
		AppendOp( op2 );
	}
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation CCodeGeneratorPSP::GenerateExitCode( u32 exit_address, u32 jump_address, u32 num_instructions, CCodeLabel next_fragment )
{
	//DAEDALUS_ASSERT( exit_address != u32( ~0 ), "Invalid exit address" );
	DAEDALUS_ASSERT( !next_fragment.IsSet() || jump_address == 0, "Shouldn't be specifying a jump address if we have a next fragment?" );

	FlushAllRegisters();

	u32		exit_delay;
	if( jump_address != 0 )
	{
		SetVar( &gCPUState.TargetPC, jump_address );
		exit_delay = EXEC_DELAY;
	}
	else
	{
		exit_delay = NO_DELAY;
	}

	LoadConstant( PspReg_A0, num_instructions );
	LoadConstant( PspReg_A1, exit_address );

	CJumpLocation	jump;

	// Use the branch delay slot to load the second argument
	PspOpCode	op1;
	PspOpCode	op2;

	GetLoadConstantOps( PspReg_A2, exit_delay, &op1, &op2 );

	if( op2._u32 == 0 )
	{
		jump = JAL( CCodeLabel( reinterpret_cast< const void * >( _DirectExitCheck ) ), false );		// No branch delay
		AppendOp( op1 );
	}
	else
	{
		AppendOp( op1 );
		jump = JAL( CCodeLabel( reinterpret_cast< const void * >( _DirectExitCheck ) ), false );		// No branch delay
		AppendOp( op2 );
	}

	
	// This jump may be NULL, in which case we patch it below
	// This gets patched with a jump to the next fragment if the target is later found
	CJumpLocation jump_to_next_fragment( J( next_fragment, true ) );

	// Patch up the exit jump if the target hasn't been compiled yet
	if( !next_fragment.IsSet() )
	{
		PatchJumpLong( jump_to_next_fragment, CCodeLabel( reinterpret_cast< const void * >( _CallReturnFromDynaRec ) ) );
	}
	return jump_to_next_fragment;
}

//*****************************************************************************
// Handle branching back to the interpreter after an ERET
//*****************************************************************************
void CCodeGeneratorPSP::GenerateEretExitCode( u32 num_instructions, CIndirectExitMap * p_map )
{
	FlushAllRegisters();

	// Eret is a bit bodged so we exit at PC + 4
	LoadConstant( PspReg_A0, num_instructions );
	LoadConstant( PspReg_A1, reinterpret_cast< s32 >( p_map ) );
	GetVar( PspReg_A2, &gCPUState.CurrentPC );
	
	J( CCodeLabel( reinterpret_cast< const void * >( _IndirectExitCheck ) ), false );
	ADDIU( PspReg_A2, PspReg_A2, 4 );		// Delay slot
}

//*****************************************************************************
// Handle branching back to the interpreter after an indirect jump
//*****************************************************************************
void CCodeGeneratorPSP::GenerateIndirectExitCode( u32 num_instructions, CIndirectExitMap * p_map )
{
	FlushAllRegisters();

	// New return address is in gCPUState.TargetPC
	LoadConstant( PspReg_A0, num_instructions );
	LoadConstant( PspReg_A1, reinterpret_cast< s32 >( p_map ) );
	GetVar( PspReg_A2, &gCPUState.TargetPC );
	
	// TODO: Use branch delay slot
	J( CCodeLabel( reinterpret_cast< const void * >( _IndirectExitCheck ) ), true );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::GenerateExceptionHander( ExceptionHandlerFn p_exception_handler_fn, std::vector< CJumpLocation > & exception_handler_jumps )
{
	CCodeLabel exception_handler( GetAssemblyBuffer()->GetLabel() );

	// XXXX - could just have J( p_fn ) here?
	JAL( CCodeLabel( reinterpret_cast< const void * >( p_exception_handler_fn ) ), true );

	J( CCodeLabel( reinterpret_cast< const void * >( _CallReturnFromDynaRec ) ), true );

	for( std::vector< CJumpLocation >::iterator it = exception_handler_jumps.begin(); it != exception_handler_jumps.end(); ++it )
	{
		CJumpLocation jump( *it );
		PatchJumpLong( jump, exception_handler );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBranchHandler( CJumpLocation branch_handler_jump, RegisterSnapshotHandle snapshot )
{
	CCodeLabel	current_label( GetAssemblyBuffer()->GetLabel() );

	PatchJumpLong( branch_handler_jump, current_label );

	mRegisterCache = GetRegisterCacheFromHandle( snapshot );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchAlways( CCodeLabel target )
{
	return J( target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfSet( const u32 * p_var, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	return BNE( PspReg_T0, PspReg_R0, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfNotSet( const u32 * p_var, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	return BEQ( PspReg_T0, PspReg_R0, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfEqual( const u32 * p_var, u32 value, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	LoadConstant( PspReg_T1, value );
	return BEQ( PspReg_T0, PspReg_T1, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfNotEqual( const u32 * p_var, u32 value, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	LoadConstant( PspReg_T1, value );
	return BNE( PspReg_T0, PspReg_T1, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetVar( u32 * p_var, EPspReg reg_src )
{
	DAEDALUS_ASSERT( reg_src != PspReg_AT, "Whoops, splattering source register" );

	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( reinterpret_cast< const u8 * >( p_var ), &reg_base, &base_offset );

	SW( reg_src, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetVar( u32 * p_var, u32 value )
{
	EPspReg		reg_value;

	if(value == 0)
	{
		reg_value = PspReg_R0;
	}
	else
	{
		LoadConstant( PspReg_T0, value );
		reg_value = PspReg_T0;
	}

	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( reinterpret_cast< const u8 * >( p_var ), &reg_base, &base_offset );

	SW( reg_value, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetFloatVar( f32 * p_var, EPspFloatReg reg_src )
{
	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( reinterpret_cast< const u8 * >( p_var ), &reg_base, &base_offset );

	SWC1( reg_src, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetVar( EPspReg dst_reg, const u32 * p_var )
{
	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( reinterpret_cast< const u8 * >( p_var ), &reg_base, &base_offset );

	LW( dst_reg, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetFloatVar( EPspFloatReg dst_reg, const f32 * p_var )
{
	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( reinterpret_cast< const u8 * >( p_var ), &reg_base, &base_offset );

	LWC1( dst_reg, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetBaseRegisterAndOffset( const u8 * p_address, EPspReg * p_reg, s16 * p_offset )
{
	s32		base_pointer_offset( reinterpret_cast< const u8 * >( p_address ) - mpBasePointer );
	if( base_pointer_offset > SHRT_MIN && base_pointer_offset < SHRT_MAX )
	{
		*p_reg = mBaseRegister;
		*p_offset = base_pointer_offset;
	}
	else
	{
		u32		address( reinterpret_cast< u32 >( p_address ) );
		u16		hi_bits( (u16)( address >> 16 ) );
		u16		lo_bits( (u16)( address ) );

		//
		//	Move up
		//
		if(lo_bits >= 0x8000)
		{
			hi_bits++;
		}

		s32		long_offset( address - ((s32)hi_bits<<16) );

		DAEDALUS_ASSERT( long_offset >= SHRT_MIN && long_offset <= SHRT_MAX, "Offset is out of range!" );

		s16		offset( (s16)long_offset );

		DAEDALUS_ASSERT( ((s32)hi_bits<<16) + offset == (s32)address, "Incorrect address calculation" );

		LUI( PspReg_AT, hi_bits );
		*p_reg = PspReg_AT;
		*p_offset = offset;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::PreEmitCachedRegCheck( const SRegisterStats & stats )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised)
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}


//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::CleanupForExit()
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBlockEnd()
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//	Generates instruction handler for the specified op code.
//	Returns a jump location if an exception handler is required
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateOpCode( u32 address, bool branch_delay_slot, OpCode op_code, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	bool	handled( false );
	bool	generate_exception_handler( false );
	bool	is_nop( op_code._u32 == 0 );

	if( is_nop )
	{
		if( branch_delay_slot )
		{
			SetVar( &gCPUState.Delay, NO_DELAY );
		}
		return CJumpLocation();
	}

	// See if this can be removed...
	if( branch_delay_slot )
	{
		SetVar( &gCPUState.Delay, EXEC_DELAY );
	}

	const EN64Reg	rs = EN64Reg( op_code.rs );
	const EN64Reg	rt = EN64Reg( op_code.rt );
	const EN64Reg	rd = EN64Reg( op_code.rd );
	const u32		sa = op_code.sa;
	const EN64Reg	base = EN64Reg( op_code.base );
	const u32		branch_target( address + ( ((s32)(s16)op_code.immediate)<<2 ) + 4);
	const u32		ft = op_code.ft;

	// 
	//	Look for opcodes we can handle manually
	//
	switch( op_code.op )
	{
	case OP_ADDI:		GenerateADDIU( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_ADDIU:		GenerateADDIU( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SLTI:		GenerateSLTI( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SLTIU:		GenerateSLTIU( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;

	case OP_ANDI:		GenerateANDI( rt, rs, op_code.immediate );			handled = true; generate_exception_handler = false; break;
	case OP_ORI:		GenerateORI( rt, rs, op_code.immediate );			handled = true; generate_exception_handler = false; break;
	case OP_XORI:		GenerateXORI( rt, rs, op_code.immediate );			handled = true; generate_exception_handler = false; break;
	case OP_LUI:		GenerateLUI( rt, s16( op_code.immediate ) );		handled = true; generate_exception_handler = false; break;

	case OP_LW:			GenerateLW( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = true; break;
	case OP_SW:			GenerateSW( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = true; break;
	case OP_LWC1:		GenerateLWC1( ft, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = true; break;
	case OP_SWC1:		GenerateSWC1( ft, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = true; break;

	case OP_BEQ:		GenerateBEQ( rs, rt, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;
	case OP_BNE:		GenerateBNE( rs, rt, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;
	case OP_BLEZ:		GenerateBLEZ( rs, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;
	case OP_BGTZ:		GenerateBGTZ( rs, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;

	case OP_SPECOP:
		switch( op_code.spec_op )
		{
		case SpecOp_SLL:	GenerateSLL( rd, rt, sa );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRA:	GenerateSRA( rd, rt, sa );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRL:	GenerateSRL( rd, rt, sa );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SLLV:	GenerateSLLV( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRAV:	GenerateSRAV( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRLV:	GenerateSRLV( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		
		case SpecOp_ADD:	GenerateADDU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_ADDU:	GenerateADDU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SUB:	GenerateSUBU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SUBU:	GenerateSUBU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		
		case SpecOp_AND:	GenerateAND( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_OR:		GenerateOR( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_XOR:	GenerateXOR( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_NOR:	GenerateNOR( rd, rs, rt );	handled = true; generate_exception_handler = false; break;

		case SpecOp_SLT:	GenerateSLT( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SLTU:	GenerateSLTU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;

		default:
			break;
		}
		break;

	case OP_COPRO1:
		switch( op_code.cop1_op )
		{
		case Cop1Op_SInstr:
			switch( op_code.cop1_funct )
			{
			case Cop1OpFunc_ADD:	GenerateADD_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_SUB:	GenerateSUB_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_MUL:	GenerateMUL_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_DIV:	GenerateDIV_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_SQRT:	GenerateSQRT_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_ABS:	GenerateABS_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_MOV:	GenerateMOV_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_NEG:	GenerateNEG_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 

			}
			break;
		default:
			break;
		}
		break;
	}

	if( !handled )
	{
		//
		//	Default handling - call interpreting function
		//
		bool	need_pc( R4300_InstructionHandlerNeedsPC( op_code ) );

		if( !need_pc )
		{
			address = 0;
		}

		GenerateGenericR4300( address, op_code, R4300_GetInstructionHandler( op_code ) );

		generate_exception_handler = need_pc;
	}

	CJumpLocation	exception_handler;
	CCodeLabel		no_target( NULL );

	if( generate_exception_handler )
	{
		// Make sure all dirty registers are flushed. NB - we don't invalidate them
		// to avoid reloading the contents if no exception was thrown.
		FlushAllRegisters();

		exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, no_target );
	}

	// Check whether we want to invert the status of this branch
	if( p_branch != NULL )
	{
		if(handled)
		{
			DAEDALUS_ASSERT( p_branch_jump->IsSet(), "Handled branch but didn't set jump" );
		}
		else
		{
			//
			// Check if the branch has been taken
			//
			if( p_branch->Direct )
			{
				if( p_branch->ConditionalBranchTaken )
				{
					*p_branch_jump = GenerateBranchIfNotEqual( &gCPUState.Delay, DO_DELAY, no_target );
				}
				else
				{
					*p_branch_jump = GenerateBranchIfEqual( &gCPUState.Delay, DO_DELAY, no_target );
				}
			}
			else
			{
				// XXXX eventually just exit here, and skip default exit code below
				if( p_branch->Eret )
				{
					*p_branch_jump = GenerateBranchAlways( no_target );
				}
				else
				{
					*p_branch_jump = GenerateBranchIfNotEqual( &gCPUState.TargetPC, p_branch->TargetAddress, no_target );
				}
			}
		}
	}
	else
	{
		if( branch_delay_slot )
		{
			SetVar( &gCPUState.Delay, NO_DELAY );
		}
	}

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateGenericR4300( u32 address, OpCode op_code, CPU_Instruction p_instruction )
{
	// Flush out the dirty registers, and mark them all as invalid
	// Theoretically we could just flush the registers referenced in the call
	FlushAndInvalidateAllRegisters();

	if( address != 0 )
	{
		SetVar( &gCPUState.CurrentPC, address );
	}

	PspOpCode	op1;
	PspOpCode	op2;

	// Load the opcode in to A0
	GetLoadConstantOps( PspReg_A0, op_code._u32, &op1, &op2 );

	if( op2._u32 == 0 )
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( p_instruction ) ), false );		// No branch delay
		AppendOp( op1 );
	}
	else
	{
		AppendOp( op1 );
		JAL( CCodeLabel( reinterpret_cast< const void * >( p_instruction ) ), false );		// No branch delay
		AppendOp( op2 );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction p_instruction )
{
	// XXXX optimise flag is ignored for now

	// Flush all fp registers before a generic call
	if (optimise)
	{
		// Flush out the dirty registers, and mark them all as invalid
		// Theoretically we could just flush the registers referenced in the call
		FlushAndInvalidateAllRegisters();
	}

	//SetVar( &gCurrentOp._u32, op_code._u32 );
	LoadConstant( PspReg_A0, op_code._u32 );
	JAL( CCodeLabel( reinterpret_cast< const void * >( p_instruction ) ), true );
}

//*****************************************************************************
//
//*****************************************************************************
static u32 Read32BitsDirect( u32 address )
{
	return *(u32 *)FuncTableReadAddress( address );
}

//*****************************************************************************
//	Read 32 bits from memory, 
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLoad32( EPspReg psp_dst, EN64Reg n64_base, s16 offset )
{
	EPspReg		reg_base( GetRegisterAndLoadLo( n64_base, PspReg_A0 ) );
	EPspReg		reg_address;
	FlushAndInvalidateAllTemporaryRegisters();
	if( offset != 0 )
	{
		reg_address = PspReg_A0;
		ADDIU( reg_address, reg_base, offset );
	}
	else
	{
		reg_address = reg_base;
	}
	SRL( PspReg_A1, reg_address, 0x12 );		// >> 18
	SLL( PspReg_V0, PspReg_A1, 0x2 );			// *4
	ADDU( PspReg_V0, PspReg_V0, PspReg_S7 );	// S7 is g_ReadAddressPointerLookupTable
	LW( PspReg_V0, PspReg_V0, 0 );				// Get table entry
	ADDU( PspReg_V0, reg_address, PspReg_V0 );	// Add offset
	CJumpLocation branch( BGEZL( PspReg_V0, CCodeLabel(NULL), false ) );
	LW( psp_dst, PspReg_V0, 0 );

	if( reg_address != PspReg_A0 )
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( Read32BitsDirect ) ), false );
		OR( PspReg_A0, reg_address, PspReg_R0 );
	}
	else
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( Read32BitsDirect ) ), true );
	}
	// Copy return value to the destination register if we're not using V0
	if( psp_dst != PspReg_V0 )
	{
		OR( psp_dst, PspReg_V0, PspReg_R0 );
	}

	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFHI( EN64Reg rd )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFLO( EN64Reg rd )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMULTU( EN64Reg rs, EN64Reg rt )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADDU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rs ]._s32[0] + gGPR[ op_code.rt ]._s32[0] );
	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	ADDU( reg_lo_d, reg_lo_a, reg_lo_b );
	UpdateRegister( rd, reg_lo_d, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUBU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rs ]._s32[0] - gGPR[ op_code.rt ]._s32[0] );
	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	SUBU( reg_lo_d, reg_lo_a, reg_lo_b );
	UpdateRegister( rd, reg_lo_d, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateAND( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = gGPR[ op_code.rs ]._u64 & gGPR[ op_code.rt ]._u64;

	if( rs == N64Reg_R0 || rt == N64Reg_R0 )
	{
		SetRegisterLo( rd, 0, PspReg_T0 );
		SetRegisterHi( rd, 0, PspReg_T0 );
	}
	else
	{
		// XXXX or into dest register
		EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
		EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
		AND( reg_lo_d, reg_lo_a, reg_lo_b );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
		EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
		AND( reg_hi_d, reg_hi_a, reg_hi_b );
		StoreRegisterHi( rd, reg_hi_d );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = gGPR[ op_code.rs ]._u64 | gGPR[ op_code.rt ]._u64;

	if( rs == N64Reg_R0 && rt == N64Reg_R0 )
	{
		SetRegisterLo( rd, 0, PspReg_T0 );
		SetRegisterHi( rd, 0, PspReg_T0 );
	}
	else if( rs == N64Reg_R0 )
	{
		// This case rarely seems to happen...
		// As RS is zero, the OR is just a copy of RT to RD.
		// Try to avoid loading into a temp register if the dest is cached
		EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		LoadRegisterLo( reg_lo_d, rt );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		LoadRegisterHi( reg_hi_d, rt );
		StoreRegisterHi( rd, reg_hi_d );
	}
	else if( rt == N64Reg_R0 )
	{
		// As RT is zero, the OR is just a copy of RS to RD.
		// Try to avoid loading into a temp register if the dest is cached
		EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		LoadRegisterLo( reg_lo_d, rs );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		LoadRegisterHi( reg_hi_d, rs );
		StoreRegisterHi( rd, reg_hi_d );
	}
	else
	{
		EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
		EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
		OR( reg_lo_d, reg_lo_a, reg_lo_b );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
		EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
		OR( reg_hi_d, reg_hi_a, reg_hi_b );
		StoreRegisterHi( rd, reg_hi_d );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateXOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = gGPR[ op_code.rs ]._u64 ^ gGPR[ op_code.rt ]._u64;

	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	XOR( reg_lo_d, reg_lo_a, reg_lo_b );
	StoreRegisterLo( rd, reg_lo_d );

	EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
	XOR( reg_hi_d, reg_hi_a, reg_hi_b );
	StoreRegisterHi( rd, reg_hi_d );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = ~(gGPR[ op_code.rs ]._u64 | gGPR[ op_code.rt ]._u64);

	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	NOR( reg_lo_d, reg_lo_a, reg_lo_b );
	StoreRegisterLo( rd, reg_lo_d );

	EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
	NOR( reg_hi_d, reg_hi_a, reg_hi_b );
	StoreRegisterHi( rd, reg_hi_d );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLT( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the registers regardless of whether we take it or not. We pull in the lo halves of 
	// the registers here so that they're Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );
	PrepareCachedRegisterLo( rt );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );

	if(rd == rs || rd == rt)
	{
		reg_lo_d = PspReg_T0;
	}

	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );

	CJumpLocation	branch( BNE( reg_hi_a, reg_hi_b, CCodeLabel(NULL), false ) );
	SLT( reg_lo_d, reg_hi_a, reg_hi_b );		// In branch delay slot

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half

	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	SLT( reg_lo_d, reg_lo_a, reg_lo_b );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rd, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the registers regardless of whether we take it or not. We pull in the lo halves of 
	// the registers here so that they're Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );
	PrepareCachedRegisterLo( rt );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );

	if(rd == rs || rd == rt)
	{
		reg_lo_d = PspReg_T0;
	}

	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );

	CJumpLocation	branch( BNE( reg_hi_a, reg_hi_b, CCodeLabel(NULL), false ) );
	SLTU( reg_lo_d, reg_hi_a, reg_hi_b );		// In branch delay slot

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half

	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	SLTU( reg_lo_d, reg_lo_a, reg_lo_b );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rd, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADDIU( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	//gGPR[op_code.rt]._s64 = (s64)(s32)(gGPR[op_code.rs]._s32[0] + (s32)(s16)op_code.immediate);

	EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
	EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	ADDIU( dst_reg, src_reg, immediate );
	UpdateRegister( rt, dst_reg, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateANDI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	//gGPR[op_code.rt]._u64 = gGPR[op_code.rs]._u64 & (u64)(u16)op_code.immediate;
	EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
	EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	ANDI( dst_reg, src_reg, immediate );
	UpdateRegister( rt, dst_reg, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateORI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	//gGPR[op_code.rt]._u64 = gGPR[op_code.rs]._u64 | (u64)(u16)op_code.immediate;

	if( rs == N64Reg_R0 )
	{
		// This may evaluate to the same code as below, but it means we can keep track of the known values and propagate
		SetRegisterLo( rt, immediate, PspReg_T0 );
		SetRegisterHi( rt, 0, PspReg_T0 );
	}
	else
	{
		EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
		EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
		ORI( dst_reg, src_reg, immediate );
		StoreRegisterLo( rt, dst_reg );

		// If the source/dest regs are different we need to copy the high bits across
		if(rt != rs)
		{
			EPspReg dst_reg_hi( GetRegisterNoLoadHi( rt, PspReg_T0 ) );
			LoadRegisterHi( dst_reg_hi, rs );
			StoreRegisterHi( rt, dst_reg_hi );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateXORI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	//gGPR[op_code.rt]._u64 = gGPR[op_code.rs]._u64 ^ (u64)(u16)op_code.immediate;
	EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
	EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	XORI( dst_reg, src_reg, immediate );
	StoreRegisterLo( rt, dst_reg );

	// If the source/dest regs are different we need to copy the high bits across
	// (if they are the same, we're xoring 0 to the top half which is essentially a NOP)
	if(rt != rs)
	{
		EPspReg dst_reg_hi( GetRegisterNoLoadHi( rt, PspReg_T0 ) );
		LoadRegisterHi( dst_reg_hi, rs );
		StoreRegisterHi( rt, dst_reg_hi );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLUI( EN64Reg rt, s16 immediate )
{
	//gGPR[op_code.rt]._s64 = (s64)(s32)((s32)(s16)op_code.immediate<<16);

	// Load into cached reg
	LUI( PspReg_T0, immediate );

	if (immediate >= 0)
	{
		UpdateRegister( rt, PspReg_T0, URO_HI_CLEAR, PspReg_T0 );
	}
	else
	{
		UpdateRegister( rt, PspReg_T0, URO_HI_SIGN_EXTEND, PspReg_T0 );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTI( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the register regardless of whether we take it or not. We pull in the lo halves of 
	// the register here so that it's Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rt, PspReg_T0 ) );

	if(rt == rs)
	{
		reg_lo_d = PspReg_T0;
	}

	CJumpLocation	branch;

	EPspReg		reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	if( immediate >= 0 )
	{
		// Positive data - we can avoid a contant load here
		branch = BNE( reg_hi_a, PspReg_R0, CCodeLabel(NULL), false );
		SLTI( reg_lo_d, reg_hi_a, 0x0000 );		// In branch delay slot
	}
	else
	{
		// Negative data
		LoadConstant( PspReg_T1, -1 );
		branch = BNE( reg_hi_a, PspReg_T1, CCodeLabel(NULL), false );
		SLTI( reg_lo_d, reg_hi_a, 0xffff );		// In branch delay slot
	}

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	SLTI( reg_lo_d, reg_lo_a, immediate );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rt, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTIU( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the register regardless of whether we take it or not. We pull in the lo halves of 
	// the register here so that it's Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rt, PspReg_T0 ) );

	if(rt == rs)
	{
		reg_lo_d = PspReg_T0;
	}

	CJumpLocation	branch;

	EPspReg		reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	if( immediate >= 0 )
	{
		// Positive data - we can avoid a contant load here
		branch = BNE( reg_hi_a, PspReg_R0, CCodeLabel(NULL), false );
		SLTIU( reg_lo_d, reg_hi_a, 0x0000 );		// In branch delay slot
	}
	else
	{
		// Negative data
		LoadConstant( PspReg_T1, -1 );
		branch = BNE( reg_hi_a, PspReg_T1, CCodeLabel(NULL), false );
		SLTIU( reg_lo_d, reg_hi_a, 0xffff );		// In branch delay slot
	}

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	SLTIU( reg_lo_d, reg_lo_a, immediate );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rt, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLL( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( (gGPR[ op_code.rt ]._u32[0] << op_code.sa) & 0xFFFFFFFF );

	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SLL( reg_lo_rd, reg_lo_rt, sa );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRL( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._u32[0] >> op_code.sa );

	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRL( reg_lo_rd, reg_lo_rt, sa );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRA( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._s32[0] >> op_code.sa );

	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRA( reg_lo_rd, reg_lo_rt, sa );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLLV( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( (gGPR[ op_code.rt ]._u32[0] << ( gGPR[ op_code.rs ]._u32[0] & 0x1F ) ) & 0xFFFFFFFF );

	// PSP sllv does exactly the same op as n64- no need for masking
	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SLLV( reg_lo_rd, reg_lo_rs, reg_lo_rt );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRLV( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._u32[0] >> ( gGPR[ op_code.rs ]._u32[0] & 0x1F ) );

	// PSP srlv does exactly the same op as n64- no need for masking
	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRLV( reg_lo_rd, reg_lo_rs, reg_lo_rt );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRAV( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._s32[0] >> ( gGPR[ op_code.rs ]._u32[0] & 0x1F ) );

	// PSP srlv does exactly the same op as n64- no need for masking
	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRAV( reg_lo_rd, reg_lo_rs, reg_lo_rt );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}


//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLB( EN64Reg rt, EN64Reg base, s32 offset )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLBU( EN64Reg rt, EN64Reg base, s32 offset )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLH( EN64Reg rt, EN64Reg base, s32 offset )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLHU( EN64Reg rt, EN64Reg base, s32 offset )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLW( EN64Reg rt, EN64Reg base, s16 offset )
{
	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//gGPR[op_code.rt]._s64 = (s64)(s32)Read32Bits(address);

	//LoadRegisterLo( PspReg_A0, base );
	//FlushAndInvalidateAllTemporaryRegisters();	
	//JAL( CCodeLabel( reinterpret_cast< const void * >( Read32Bits ) ), false );		// Don't set branch delay slot
	//ADDIU( PspReg_A0, PspReg_A0, offset );

	EPspReg	reg_dst( GetRegisterNoLoadLo( rt, PspReg_V0 ) );	// Use V0 to avoid copying return value if reg is not cached

	GenerateLoad32( reg_dst, base, offset );

	// XXXX we should check exception handler here, BEFORE setting result

	UpdateRegister( rt, reg_dst, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLWC1( u32 ft, EN64Reg base, s32 offset )
{
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EPspFloatReg	psp_ft( GetFloatRegisterNoLoad( n64_ft ) );

	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//value = Read32Bits(address);
	//g_qwCPR[1][op_code.ft]._s32[ 0 ] = value;

	// TODO: Actually perform LWC1 here
	EPspReg	reg_dst( PspReg_V0 );				// GenerateLoad32 is slightly more efficient when using V0
	GenerateLoad32( reg_dst, base, offset );

	//SetVar( &g_qwCPR[1][ ft ]._u32[ 0 ], reg_dst );
	MTC1( psp_ft, reg_dst );

	UpdateFloatRegister( n64_ft );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSB( EN64Reg rt, EN64Reg base, s32 offset )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSH( EN64Reg rt, EN64Reg base, s32 offset )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSW( EN64Reg rt, EN64Reg base, s32 offset )
{
	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//Write32Bits(address, gGPR[op_code.rt]._u32[0]);

	LoadRegisterLo( PspReg_A0, base );
	LoadRegisterLo( PspReg_A1, rt );

	FlushAndInvalidateAllTemporaryRegisters();
	JAL( CCodeLabel( reinterpret_cast< const void * >( Write32Bits ) ), false );		// Don't set branch delay slot
	ADDIU( PspReg_A0, PspReg_A0, offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSWC1( u32 ft, EN64Reg base, s32 offset )
{
	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//Write32Bits(address, g_qwCPR[1][op_code.ft]._s32[ 0 ] );

	LoadRegisterLo( PspReg_A0, base );
	//GetVar( PspReg_A1, &g_qwCPR[1][ft]._u32[ 0 ] );

	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );
	MFC1( PspReg_A1, psp_ft );

	FlushAndInvalidateAllTemporaryRegisters();
	JAL( CCodeLabel( reinterpret_cast< const void * >( Write32Bits ) ), false );		// Don't set branch delay slot
	ADDIU( PspReg_A0, PspReg_A0, offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFC1( EN64Reg rt, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMTC1( u32 fs, EN64Reg rt )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCFC1( EN64Reg rt, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCTC1( u32 fs, EN64Reg rt )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBEQ( EN64Reg rs, EN64Reg rt, u32 target )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBEQ( EN64Reg rs, EN64Reg rt, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BEQ?" );

	// One or other of these may be r0 - we don't really care for optimisation purposes though
	// as ultimately the register load regs factored out
	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg		reg_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BNE( reg_a, reg_b, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BEQ( reg_a, reg_b, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBNE( EN64Reg rs, EN64Reg rt, u32 target )
{
	NOT_IMPLEMENTED( __FUNCTION__ );

	// Legacy
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBNE( EN64Reg rs, EN64Reg rt, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BNE?" );

	// One or other of these may be r0 - we don't really care for optimisation purposes though
	// as ultimately the register load regs factored out
	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg		reg_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BEQ( reg_a, reg_b, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BNE( reg_a, reg_b, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBLEZ( EN64Reg rs, u32 target )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBLEZ( EN64Reg rs, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BLEZ?" );

	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BGTZ( reg_a, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BLEZ( reg_a, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBGTZ( EN64Reg rs, u32 target )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBGTZ( EN64Reg rs, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BGTZ?" );

	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BLEZ( reg_a, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BGTZ( reg_a, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADD_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	ADD_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUB_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	SUB_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMUL_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	MUL_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	DIV_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSQRT_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	SQRT_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateABS_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	ABS_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMOV_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	MOV_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNEG_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	NEG_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCompare_S( u32 fs, u32 ft, u16 flag )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADD_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUB_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMUL_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSQRT_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMOV_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNEG_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCompare_D( u32 fs, u32 ft, u16 flag )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_D_S( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_W_S( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_D_W( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_S_W( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_S_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

