/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "FragmentCache.h"

#include "Fragment.h"
#include "CodeBufferManager.h"

#include "Debug/DBGConsole.h"

#include "Utility/Profiler.h"

#include "AssemblyUtils.h"

#include <algorithm>

using namespace AssemblyUtils;

//*************************************************************************************
//
//*************************************************************************************
CFragmentCache::CFragmentCache()
:	mMemoryUsage( 0 )
,	mInputLength( 0 )
,	mOutputLength( 0 )
,	mCachedFragmentAddress( 0 )
,	mpCachedFragment( NULL )
{
	mpCodeBufferManager = CCodeBufferManager::Create();
	if(mpCodeBufferManager != NULL)
	{
		mpCodeBufferManager->Initialise();
	}
}

//*************************************************************************************
//
//*************************************************************************************
CFragmentCache::~CFragmentCache()
{
	Clear();

	mpCodeBufferManager->Finalise();
	delete mpCodeBufferManager;
}

//*************************************************************************************
//
//*************************************************************************************
CFragment * CFragmentCache::LookupFragment( u32 address ) const
{
	DAEDALUS_PROFILE( "CFragmentCache::LookupFragment" );

	if( address != mCachedFragmentAddress )
	{
		mCachedFragmentAddress = address;

		FragmentMap::const_iterator	fragment( mFragmentMap.find( address ) );
		if(fragment != mFragmentMap.end())
		{
			mpCachedFragment = fragment->second;
		}
		else
		{
			mpCachedFragment = NULL;
		}
	}

	return mpCachedFragment;
}

//*************************************************************************************
//
//*************************************************************************************
void CFragmentCache::InsertFragment( CFragment * p_fragment )
{
	u32		fragment_address( p_fragment->GetEntryAddress() );

	DAEDALUS_ASSERT( mFragmentMap.find( fragment_address ) == mFragmentMap.end(), "A fragment with this address already exists" );
	mFragmentMap.insert( FragmentMap::value_type( fragment_address, p_fragment ) );

	// Process any jumps for this before inserting new ones
	JumpMap::iterator	jump_it( mJumpMap.find( fragment_address ) );
	if( jump_it != mJumpMap.end() )
	{
		const JumpList &		jumps( jump_it->second );
		for( JumpList::const_iterator it = jumps.begin(); it != jumps.end(); ++it )
		{
			//DBGConsole_Msg( 0, "Inserting [R%08x], patching jump at %08x ", address, (*it) );
			PatchJumpLong( (*it), p_fragment->GetEntryTarget() );
		}	 

		// All patched - clear
		mJumpMap.erase( jump_it );
	}

	// Finally register any links that this fragment may have
	const FragmentPatchList &	patch_list( p_fragment->GetPatchList() );
	for( FragmentPatchList::const_iterator it = patch_list.begin(); it != patch_list.end(); ++it )
	{
		u32				target_address( it->Address );
		CJumpLocation	jump( it->Jump );

		if( jump.IsSet() )
		{
			CFragment * p_fragment( LookupFragment( target_address ) );
			if( p_fragment != NULL )
			{
				PatchJumpLong( jump, p_fragment->GetEntryTarget() );

				DAEDALUS_ASSERT( mJumpMap.find( target_address ) == mJumpMap.end(), "Jump map still contains an entry for this" );
			}
			else if( target_address != u32(~0) )
			{
				// Store the address for later processing
				mJumpMap[ target_address ].push_back( jump );
			}
		}
		else
		{
			DAEDALUS_ASSERTMSG( "No exit jump?" );
		}
	}

	// Free memoire
	p_fragment->DiscardPatchList();

	// For simulation only
	p_fragment->SetCache( this );

	// Update memory usage etc after discarding patch list
	mMemoryUsage += p_fragment->GetMemoryUsage();
	mInputLength += p_fragment->GetInputLength();
	mOutputLength += p_fragment->GetOutputLength();

	if((mFragmentMap.size() % 100) == 0)
	{
		u32		expansion = 1;
		if(mInputLength > 0)
		{
			expansion = (100 * mOutputLength) / mInputLength;
		}
#ifdef DAEDALUS_PSP
		printf( "Dynarec: %d fragments, %dKB (i:o %d:%d = %d%%)\n",
			mFragmentMap.size(), mMemoryUsage / 1024, mInputLength / 1024, mOutputLength / 1024, expansion );
#else
		DBGConsole_Msg( 0, "Dynarec: %d fragments, %dKB (i:o %d:%d = %d%%)",
			mFragmentMap.size(), mMemoryUsage / 1024, mInputLength / 1024, mOutputLength / 1024, expansion );
#endif
	}
}

//*************************************************************************************
//
//*************************************************************************************
void CFragmentCache::Clear()
{
	// Clear out all the framents
	for(FragmentMap::iterator it = mFragmentMap.begin(); it != mFragmentMap.end(); ++it)
	{
		delete it->second;
	}

	mFragmentMap.clear();
	mMemoryUsage = 0;
	mInputLength = 0;
	mOutputLength = 0;
	mCachedFragmentAddress = 0;
	mpCachedFragment = NULL;
	mJumpMap.clear();

	mpCodeBufferManager->Reset();
}

#ifdef _DEBUG
//*************************************************************************************
//
//*************************************************************************************
struct SDescendingHitCountSort
{
     bool operator()(CFragment* const & a, CFragment* const & b)
     {
          return b->GetHitCount() < a->GetHitCount();
     }
};

//*************************************************************************************
//
//*************************************************************************************
void CFragmentCache::DumpStats( const char * filename ) const
{
	typedef std::vector< CFragment * >		FragmentList;
	FragmentList		all_fragments;

	all_fragments.reserve( mFragmentMap.size() );

	// Sort in order of hit count
	for(FragmentMap::const_iterator it = mFragmentMap.begin(); it != mFragmentMap.end(); ++it)
	{
		all_fragments.push_back( it->second );
	}

	std::sort( all_fragments.begin(), all_fragments.end(), SDescendingHitCountSort() );


	// Dump out the trace info
	FILE * fh( fopen( filename, "w" ) );
	if(fh)
	{
		for(FragmentList::const_iterator it = all_fragments.begin(); it != all_fragments.end(); ++it)
		{
			CFragment * p_fragment( *it );
			p_fragment->DumpFragmentInfo( fh );
		}

		fclose( fh );
	}
}
#endif // _DEBUG
