/*
Copyright (C) 2001 CyRUS64 (http://www.boob.co.uk)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "IniFile.h"

#include "Core/ROM.h"
#include "RomDB.h"

#include "DaedIO.h"

#include <string>
#include <vector>

//*****************************************************************************
//
//*****************************************************************************
class IIniFile : public CIniFile
{
	public:
		IIniFile();
		virtual ~IIniFile();

		//
		// CIniFile implementation
		//
		bool			OpenIniFile( const CHAR * filename );
		void			Commit();												// (STRMNNRMN - Write ini back out to disk?)

		bool			GetSettings( const RomID & id, RomSettings * p_settings ) const;
		void			SetSettings( const RomID & id, const RomSettings & settings );

	private:

		//
		// Utility functions
		//
		int				FindEntry( const RomID & id ) const; 		// Find entry in list
		void			AddEntry( const RomID & id, const RomSettings & settings );

		void			OutputSectionDetails(u32 i, FILE * fh);

	private:

		struct section
		{
			RomID				mRomID;
			RomSettings			mSettings;
			bool				bOutput;			// (STRMNNRMN - Has it been written to disk yet?)
		};

		std::vector<section>	mSections;

		bool					mDirty;				// (STRMNNRMN - Changed since read from disk?)
		TCHAR					mFileName[ MAX_PATH + 1 ];
};



//*****************************************************************************
// Singleton creator
//*****************************************************************************
template<> bool	CSingleton< CIniFile >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new IIniFile();

	return true;
};


//*****************************************************************************
// Constructor
//*****************************************************************************
IIniFile::IIniFile()
{
	mSections.clear();
	mDirty = false;
	mFileName[ 0 ] = '\0';
}

//*****************************************************************************
//
//*****************************************************************************
IIniFile::~IIniFile()
{
	if ( mDirty )
	{
		Commit();
	}
}

//*****************************************************************************
// (STRMNNRMN - Strip spaces from end of names)
//*****************************************************************************
static char * tidy(char * s)
{
	char * p = s + strlen(s);

	p--;
	while (p >= s && (*p == ' ' || *p == '\r' || *p == '\n'))
	{
		*p = 0;
		p--;
	}
	return s;

}

//*****************************************************************************
//	Remove the specified characters from p_string
//*****************************************************************************
static bool	trim( char * p_string, const char * p_trim_chars )
{
	u32 num_trims = strlen( p_trim_chars );
	char * pin = p_string;
	char * pout = p_string;
	bool found;
	while ( *pin )
	{
		char c = *pin;

		found = false;
		for ( u32 i = 0; i < num_trims; i++ )
		{
			if ( p_trim_chars[ i ] == c )
			{
				// Skip
				found = true;
				break;
			}
		}

		if ( found )
		{
			pin++;
		}
		else
		{
			// Copy
			*pout++ = *pin++;
		}
	}
	*pout = '\0';
	return true;
}

//*****************************************************************************
//
//*****************************************************************************
static bool	StringEquals( const char * p_a, const char * p_b )
{
	return stricmp( p_a, p_b ) == 0;
}

//*****************************************************************************
//
//*****************************************************************************
static RomID	RomIDFromString( const CHAR * str )
{
	u32 crc1, crc2, country;
	sscanf( str, "%08x%08x-%02x", &crc1, &crc2, &country );
	return RomID( crc1, crc2, (u8)country );
}

//*****************************************************************************
//
//*****************************************************************************
bool IIniFile::OpenIniFile( const CHAR * filename )
{
	const u32	BUFFER_LEN = 128;
	char readinfo[BUFFER_LEN+1];
	char trim_chars[]="{}"; //remove first and last character

	//
	// Remember the filename
	//
	strcpy( mFileName, filename );
	
	FILE * fh( fopen( filename, "r" ) );
	if (fh == NULL)
		return false;

	while (fgets( readinfo, BUFFER_LEN, fh ) != NULL)
	{
		tidy(readinfo);			// STRMNNRMN - Strip spaces from end of lines

		// STRMNNRMN - comments
		if (readinfo[0] == '/')
			continue;

		if (!stricmp(readinfo,"")==0)
		{
			if (readinfo[0] == '{') //if a section heading
			{
				section newsection;

				trim(readinfo,trim_chars);

				newsection.mRomID = RomIDFromString( readinfo );
				//strcpy(newsection.crccheck, readinfo);
				newsection.bOutput = false;
				ROM_ResetSettings( &newsection.mSettings );

				// Temp hack

				mSections.push_back(newsection);
				
			}
			else
			{		
				int sectionno = mSections.size() - 1;

				std::string	str( readinfo );
				std::string	key, value;

				int		equals_idx = str.find( '=', 0 );
				if( equals_idx >= 0 )
				{
					key = str.substr( 0, equals_idx );
					value = str.substr( equals_idx+1 );
				}
				else
				{
					key = str;
				}
			
				if(StringEquals( key.c_str(), "Comment" ))
				{
					strcpy(mSections[sectionno].mSettings.szComment, value.c_str());
				}
				else if(StringEquals( key.c_str(), "Info" ))
				{
					strcpy(mSections[sectionno].mSettings.szInfo, value.c_str());
				}
				else if(StringEquals( key.c_str(), "Name" ))
				{
					strcpy(mSections[sectionno].mSettings.szGameName, value.c_str());
				}
				else if(StringEquals( key.c_str(), "DisableDynarec" ))
				{
					mSections[sectionno].mSettings.DynarecEnabled = false;
				}
				else if(StringEquals( key.c_str(), "DisablePatches" ))
				{
					mSections[sectionno].mSettings.PatchesEnabled = false;
				}
				else if(StringEquals( key.c_str(), "DisableTextureCRC" ))
				{
					mSections[sectionno].mSettings.TextureCRCEnabled = false;
				}
				else if(StringEquals( key.c_str(), "DisableSpeedSync" ))
				{
					mSections[sectionno].mSettings.SpeedSyncEnabled = false;
				}
				else if(StringEquals( key.c_str(), "ExpansionPak" ))
				{
					mSections[sectionno].mSettings.ExpansionPakEnabled = true;
				}
				else if(StringEquals( key.c_str(),"SaveEep4k" ))
				{
					mSections[sectionno].mSettings.SaveType = SAVE_TYPE_EEP4K;
				}
				else if(StringEquals( key.c_str(),"SaveEep16k" ))
				{
					mSections[sectionno].mSettings.SaveType = SAVE_TYPE_EEP16K;
				}
				else if(StringEquals( key.c_str(),"SaveSRAM" ))
				{
					mSections[sectionno].mSettings.SaveType = SAVE_TYPE_SRAM;
				}
				else if(StringEquals( key.c_str(),"SaveFlashRam" ))
				{
					mSections[sectionno].mSettings.SaveType = SAVE_TYPE_FLASH;
				}

				// No longer used
				//if (StringEquals(key.c_str(), "OSRescanCount"))
				//	mSections[sectionno].mSettings.dwRescanCount = strtol(right(key.c_str(),4),NULL,10);
			}
		}
	}
	fclose(fh);
	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void IIniFile::Commit()
{
	TCHAR szFileNameOut[MAX_PATH+1];
	TCHAR szFileNameDelete[MAX_PATH+1];
	u32 i;
	FILE * fhIn;
	FILE * fhOut;
	TCHAR szBuf[1024+1];
	char trim_chars[]="{}\n\r"; //remove first and last character

	sprintf(szFileNameOut, "%s.tmp", mFileName);
	sprintf(szFileNameDelete, "%s.del", mFileName);

	fhIn = fopen(mFileName, "r");
	if (fhIn == NULL)
		return;

	fhOut = fopen(szFileNameOut, "w");
	if (fhOut == NULL)
	{
		fclose(fhIn);
		return;
	}

	// Mark all mSections and needing to be written
	for (i = 0; i < mSections.size(); i++)
	{
		mSections[i].bOutput = false;
	}


	while (fgets(szBuf, 1024, fhIn))
	{
		if (szBuf[0] == '{')
		{
			BOOL bFound = FALSE;
			// Start of section
			trim(szBuf,trim_chars);

			RomID id = RomIDFromString( szBuf );

			for (i = 0; i < mSections.size(); i++)
			{
				if (mSections[i].bOutput)
					continue;

				if ( mSections[i].mRomID == id )
				{
					// Output this CRC
					OutputSectionDetails(i, fhOut);
					mSections[i].bOutput = true;
					bFound = TRUE;
					break;
				}
			}
			if (!bFound)
			{
				// Do what? This should never happen, unless the user
				// replaces the inifile while Daedalus is running!
			}
		}
		else if (szBuf[0] == '/')
		{
			// Comment
			fputs(szBuf, fhOut);
			continue;
		}

	}

	// Input buffer done-  process any new entries!
	for (i = 0; i < mSections.size(); i++)
	{
		// Skip any that have not been done.
		if (mSections[i].bOutput)
			continue;
		// Output this CRC
		// Removed at request of Genueix :)
		//fprintf(fhOut, "// Automatically generated entry - may need editing\n");
		OutputSectionDetails(i, fhOut);
		mSections[i].bOutput = true;
	}

	fclose(fhOut);
	fclose(fhIn);

	// Create the new file
	IO::File::Move(mFileName, szFileNameDelete);
	IO::File::Move(szFileNameOut, mFileName);
	IO::File::Delete(szFileNameDelete);

	mDirty = false;
}


//*****************************************************************************
//
//*****************************************************************************
void IIniFile::OutputSectionDetails(u32 i, FILE * fh)
{
	// Generate the CRC-ID for this rom:
	fprintf(fh, "{%08x%08x-%02x}\n", mSections[i].mRomID.CRC[0], mSections[i].mRomID.CRC[1], mSections[i].mRomID.CountryID );
	
	fprintf(fh, "Name=%s\n", mSections[i].mSettings.szGameName);

	if (strlen(mSections[i].mSettings.szComment) > 0)	fprintf(fh, "Comment=%s\n", mSections[i].mSettings.szComment);
	if (strlen(mSections[i].mSettings.szInfo) > 0)		fprintf(fh, "Info=%s\n", mSections[i].mSettings.szInfo);
	if (!mSections[i].mSettings.DynarecEnabled)			fprintf(fh, "DisableDynarec\n");
	if (!mSections[i].mSettings.PatchesEnabled)			fprintf(fh, "DisablePatches\n");
	if (!mSections[i].mSettings.TextureCRCEnabled)		fprintf(fh, "DisableTextureCRC\n");
	if (!mSections[i].mSettings.SpeedSyncEnabled)		fprintf(fh, "DisableSpeedSync\n");
	if (mSections[i].mSettings.ExpansionPakEnabled)		fprintf(fh, "ExpansionPak\n");
	if (mSections[i].mSettings.dwRescanCount != 0)		fprintf(fh, "OSRescanCount=%d\n", mSections[i].mSettings.dwRescanCount);

	switch ( mSections[i].mSettings.SaveType )
	{
	case SAVE_TYPE_UNKNOWN:		break;									// Print nothing
	case SAVE_TYPE_EEP4K:		fprintf(fh, "SaveEep4k\n" ); break;
	case SAVE_TYPE_EEP16K:		fprintf(fh, "SaveEep16k\n" ); break;
	case SAVE_TYPE_SRAM:		fprintf(fh, "SaveSRAM\n" ); break;
	case SAVE_TYPE_FLASH:		fprintf(fh, "SaveFlashRam\n" ); break;
	}

	fprintf(fh, "\n");			// Spacer
}

//*****************************************************************************
// Internal function to add a new entry to the database
//*****************************************************************************
void IIniFile::AddEntry( const RomID & id, const RomSettings & settings )
{
	// Add new entry!!!
	section newsection;

	newsection.mRomID = id;
	newsection.bOutput = false;
	newsection.mSettings = settings;

	mSections.push_back(newsection);

	mDirty = true;				// Flag to indicate we should be updated
}

//*****************************************************************************
// Internal function to find the entry corresponding to the specified rom. 
//*****************************************************************************
int IIniFile::FindEntry( const RomID & id ) const
{
	for ( u32 i = 0; i < mSections.size(); i++)
	{
		if ( mSections[i].mRomID == id )
			return i;
	}

	//
	// Not found
	//
	return -1;
}

//*****************************************************************************
// Retreive the settings for the specified rom. Returns false if the rom is 
// not in the database
//*****************************************************************************
bool	IIniFile::GetSettings( const RomID & id, RomSettings * p_settings ) const
{
	int i = FindEntry( id );
	if ( i >= 0 )
	{
		*p_settings = mSections[ i ].mSettings;
		return true;
	}
	else
	{
		return false;
	}
}

//*****************************************************************************
// Update the settings for the specified rom - creates a new entry if necessary
//*****************************************************************************
void	IIniFile::SetSettings( const RomID & id, const RomSettings & settings )
{
	int i = FindEntry( id );
	if ( i >= 0 )
	{
		mSections[ i ].mSettings = settings;
	}
	else
	{
		AddEntry( id, settings );
	}
	
	mDirty = true;
}

