/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "Texture.h"
#include "NativeTexture.h"
#include "PngUtil.h"
#include "DaedMathUtil.h"

#include "Debug\DBGConsole.h"

#include <pspgu.h>
#include <psputils.h>

//*****************************************************************************
//
//*****************************************************************************
class ITexture : public CTexture
{
	public:
		ITexture();
		virtual ~ITexture();

		bool					Initialise(u32 width, u32 height);
		CNativeTexture *		GetTexture()					{ return mpTexture; }
		CNativeTexture *		GetRecolouredTexture( u32 mask, u32 colour );

#ifdef LIBPNG_SUPPORT
		bool					DumpImageAsPNG(const char * filename, u32 width, u32 height);
#endif
		virtual u32				GetPixelSize() const;

		virtual	u32				GetWidth() const				{ return mCorrectedWidth; }
		virtual	u32				GetHeight() const				{ return mCorrectedHeight; }

		virtual u32				GetVideoMemoryUsage() const;
		virtual u32				GetSystemMemoryUsage() const;

		// Provides access to "surface"
		bool					StartUpdate(DrawInfo *di);
		void					EndUpdate(DrawInfo *di);

	protected:

		u8 *					mpTexels;
		CNativeTexture *		mpTexture;
		CNativeTexture *		mpRecolouredTexture;
		
		u32						mWidth;			// The requested Texture w/h
		u32						mHeight;

		u32						mCorrectedWidth;	// What was actually created
		u32						mCorrectedHeight;
};



//*****************************************************************************
//
//*****************************************************************************
CTexture * CTexture::Create( u32 width, u32 height )
{
	ITexture * pTexture = new ITexture();

	if (!pTexture->Initialise(width, height))
	{
		delete pTexture;
		return NULL;
	}

	return pTexture;
}

//*****************************************************************************
//
//*****************************************************************************
ITexture::ITexture() :
	mpTexels(NULL),
	mpTexture(NULL),
	mpRecolouredTexture(NULL),
	mWidth(0),
	mHeight(0),
	mCorrectedWidth(0),
	mCorrectedHeight(0)
{
}

//*****************************************************************************
//
//*****************************************************************************
ITexture::~ITexture(void)
{
	delete [] mpTexels;

	if (mpTexture)
	{
		mpTexture->Release();
	}
	if(mpRecolouredTexture)
	{
		mpRecolouredTexture->Release();
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool ITexture::Initialise(u32 width, u32 height)
{
	DAEDALUS_ASSERT_Q(mpTexture == NULL);

	CNativeTexture::CorrectDimensions( width, height, mCorrectedWidth, mCorrectedHeight );

	mpTexture = new CNativeTexture( mCorrectedWidth, mCorrectedHeight );

	if(mpTexture != NULL)
	{
		u32		buffer_size( mCorrectedWidth * mCorrectedHeight * GetPixelSize() );

		mpTexels = new u8[ buffer_size ];

		mWidth = width;
		mHeight = height;
		return true;
	}

	return false;
}

//*****************************************************************************
//
//*****************************************************************************
u32 ITexture::GetPixelSize() const
{
	return 4;
}
//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	ITexture::GetRecolouredTexture( u32 mask, u32 colour )
{
	if(mpRecolouredTexture == NULL)
	{
		mpRecolouredTexture = new CNativeTexture( mCorrectedWidth, mCorrectedHeight );

		if(mpRecolouredTexture != NULL)
		{
			mpRecolouredTexture->SetDataForceColour( mpTexels, mCorrectedWidth, mCorrectedHeight, mask, colour );
		}
	}

	return mpRecolouredTexture;
}

//*****************************************************************************
// Lock the texture so we can start updating it
//*****************************************************************************
bool ITexture::StartUpdate(DrawInfo *di)
{
	if (mpTexture != NULL)
	{
		if(mpTexels != NULL)
		{
			// Return a temporary buffer to use
			di->pSurface = mpTexels;
			di->Width = mCorrectedWidth;
			di->Height = mCorrectedHeight;
			di->Pitch = mCorrectedWidth * GetPixelSize();

			return true;
		}
	}

	return false;
}

//*****************************************************************************
// Finish updating the texture and copy the memory to the native surface
//*****************************************************************************
void ITexture::EndUpdate(DrawInfo *di)
{
	if (mpTexture != NULL)
	{
		mpTexture->SetData( mpTexels, mCorrectedWidth, mCorrectedHeight );
	}
}

//*****************************************************************************
//
//*****************************************************************************
u32	ITexture::GetVideoMemoryUsage() const
{
	u32		usage( 0 );

	if(mpTexture != NULL)
	{
		usage += mpTexture->GetVideoMemoryUsage();
	}

	if(mpRecolouredTexture != NULL)
	{
		usage += mpRecolouredTexture->GetVideoMemoryUsage();
	}

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
u32	ITexture::GetSystemMemoryUsage() const
{
	u32	usage( mCorrectedWidth * mCorrectedHeight * GetPixelSize() );		// mpTexels

	if(mpTexture != NULL)
	{
		usage += mpTexture->GetSystemMemoryUsage();
	}

	if(mpRecolouredTexture != NULL)
	{
		usage += mpRecolouredTexture->GetSystemMemoryUsage();
	}

	return usage;
}

#ifdef LIBPNG_SUPPORT
//*****************************************************************************
//
//*****************************************************************************
bool ITexture::DumpImageAsPNG(const char * filename, u32 width, u32 height)
{
	u32 * pB = new u32 [width * height];
	if (pB == NULL)
	{
		return false;
	}

	DrawInfo di;
	
	if (!StartUpdate(&di))
	{
		delete [] pB;
		return false;
	}
	
	u32 *pSrc = (u32 *)di.lpSurface;
	LONG lPitch = di.lPitch;

	u32 index = 0;

	for (u32 y = 0; y < height; y++)
	{
		u32 * dwSrc = (u32 *)((u8 *)pSrc + y*lPitch);
		for (u32 x = 0; x < width; x++)
		{
			pB[index++] = dwSrc[x];
		}
	}
	EndUpdate(&di);
	
	PngSaveImage(filename, (png_byte *)pB, width, height);

	delete [] pB;

	return true;
}

#endif // LIBPNG_SUPPORT
