/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

// Manages textures for RDP code
// Uses a HashTable (hashing on TImg) to allow quick access
//  to previously used textures

#include "stdafx.h"

//#include "memory.h"
#include "TextureCache.h"
#include "ConvertImage.h"
#include "RDP.h"

#include "ultra_gbi.h"		// For image size/formats
#include "Utility/Profiler.h"
#include "Core/ROM.h"
#include "Core/Memory.h"

#include "Debug/DaedalusAssert.h"
#include "Debug/DBGConsole.h"

#include "DaedHash.h"

#include <pspdebug.h>

extern u32 gRDPFrame;

// Returns the first prime greater than or equal to nFirst
static s32 GetNextPrime(s32 nFirst)
{
	s32 nCurrent;
	
	s32 i;
	
	nCurrent = nFirst;
	
	// Just make sure it's odd
	if ((nCurrent % 2) == 0)
		nCurrent++;
	
	for (;;)
	{
		s32 nSqrtCurrent;
		BOOL bIsComposite;
		
		// nSqrtCurrent = nCurrent^0.5 + 1 (round up)
		nSqrtCurrent = (s32)sqrt((double)nCurrent) + 1;
		
		bIsComposite = FALSE;
		
		// Test all odd numbers from 3..nSqrtCurrent
		for (i = 3; i <= nSqrtCurrent; i+=2)
		{
			if ((nCurrent % i) == 0)
			{
				bIsComposite = TRUE;
				break;
			}
		}
		
		if (!bIsComposite)
		{			
			return nCurrent;
		}
		
		// Select next odd candidate...
		nCurrent += 2;
	}
	
}


// Implementation

typedef struct TextureEntry
{
	TextureEntry() : pTexture(NULL), pMirroredTexture(NULL) {}

	~TextureEntry()
	{
		if (pTexture)
		{
			pTexture->Release();
			pTexture = NULL;
		}
		if (pMirroredTexture)
		{
			pMirroredTexture->Release();
			pMirroredTexture = NULL;
		}
	}

	u32		GetVideoMemoryUsage() const
	{
		u32	usage( 0 );
		if( pTexture != NULL )
		{
			usage += pTexture->GetVideoMemoryUsage();
		}
		if( pMirroredTexture != NULL )
		{
			usage += pMirroredTexture->GetVideoMemoryUsage();
		}
		return usage;
	}

	u32		GetSystemMemoryUsage() const
	{
		u32	usage( 0 );
		if( pTexture != NULL )
		{
			usage += pTexture->GetSystemMemoryUsage();
		}
		if( pMirroredTexture != NULL )
		{
			usage += pMirroredTexture->GetSystemMemoryUsage();
		}
		return usage;
	}


	//
	// Data
	//
	struct TextureEntry * pNext;

	TextureInfo ti;
	u32			HashValue;

	u32			NumUses;			// Total times used (for stats)
	u32			FrameLastUsed;	// Frame # that this was last used

	CTexture	* pTexture;
	// TODO: when nXTimes and nYTimes in the mirror emulator are correctly determined, these should be put here
	CTexture * pMirroredTexture;

} TextureEntry;



class ITextureCache : public CTextureCache
{
public:
	ITextureCache();
	~ITextureCache();
	
			bool			Initialise(u32 size);

			void			PurgeOldTextures();
			u32				CountOldTextures() const;
			void			DropTextures();

			void			SetDumpTextures( bool dump_textures )	{ mDumpTextures = dump_textures; }
			bool			GetDumpTextures( ) const				{ return mDumpTextures; }


			bool			InvalidateDeviceObjects();
	
			CTexture *		GetTexture(TextureInfo * pti);
	
	virtual	void			DisplayStats();

protected:
			TextureEntry *	CreateEntry(TextureInfo * pti);
			void			AddTextureEntry(TextureEntry *p_entry);
			void			RemoveTextureEntry(TextureEntry * p_entry);
			
			void			AddToRecycleList(TextureEntry *p_entry);
			TextureEntry *	ReviveUsedTexture(u32 width, u32 height);
			
			
			TextureEntry *	GetEntry(TextureInfo * pti);
			TextureEntry *	GetEntryAndUpdate(TextureInfo * pti);
			
			u32				GetHashIndex(u32 value);
			
			void			DecompressTexture(TextureEntry * p_entry);
			void			DecompressTmemTexture(TextureEntry * p_entry);

			u32				GenerateHashValue(TextureInfo * pti) const;

			void			GetUsedTextureStats( u32 * p_num_textures, u32 * p_video_memory_used, u32 * p_system_memory_used ) const;
			void			GetOldTextureStats( u32 * p_num_textures, u32 * p_video_memory_used, u32 * p_system_memory_used ) const;

protected:
	TextureEntry *			mpFirstUsedSurface;
	TextureEntry **			mpTextureHashTable;
	u32						mTextureHashSize;
	bool					mDumpTextures;
	
};


// Interface

//*****************************************************************************
//
//*****************************************************************************

template<> bool CSingleton< CTextureCache >::Create()
{
	ITextureCache * pNewInstance;
	
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	pNewInstance = new ITextureCache();
	if (!pNewInstance)
	{
		return false;
	}
	else
	{
		if (!pNewInstance->Initialise(8000))
		{
			delete pNewInstance;
			pNewInstance = NULL;
		}
	}
	
	mpInstance = pNewInstance;
	return true;
}

//*************************************************************************************
//
//*************************************************************************************
ITextureCache::ITextureCache() :
	mpFirstUsedSurface(NULL),
	mpTextureHashTable(NULL),
	mTextureHashSize(0),
	mDumpTextures(false)
{
}

//*************************************************************************************
//
//*************************************************************************************
ITextureCache::~ITextureCache()
{
	DropTextures();
	
	while (mpFirstUsedSurface)
	{
		TextureEntry * pVictim = mpFirstUsedSurface;
		mpFirstUsedSurface = pVictim->pNext;
		
		delete pVictim;
	}

	delete []mpTextureHashTable;
	mpTextureHashTable = NULL;	
}

//*************************************************************************************
// Called when the device-dependent objects are about to be lost.
//*************************************************************************************
bool ITextureCache::InvalidateDeviceObjects()
{
	DropTextures();
	
	while (mpFirstUsedSurface)
	{
		TextureEntry * pVictim = mpFirstUsedSurface;
		mpFirstUsedSurface = pVictim->pNext;
		
		delete pVictim;
	}

	return true;
}

//*************************************************************************************
//
//*************************************************************************************
bool ITextureCache::Initialise(u32 size)
{
	size = GetNextPrime(size);
	
	mpTextureHashTable = new TextureEntry *[size];
	if (mpTextureHashTable == NULL)
		return false;
	
	mTextureHashSize = size;
	
	for( u32 i = 0; i < size; i++ )
		mpTextureHashTable[i] = NULL;
	
	return true;	// Success!
	
}

//*************************************************************************************
//
//*************************************************************************************
void	ITextureCache::GetUsedTextureStats( u32 * p_num_textures,
											u32 * p_video_memory_used,
											u32 * p_system_memory_used ) const
{
	u32	num_textures( 0 );
	u32	video_memory_used( 0 );
	u32	system_memory_used( 0 );

	for( u32 i = 0; i < mTextureHashSize; ++i )
	{
		for( TextureEntry * p_te = mpTextureHashTable[ i ]; p_te != NULL; p_te = p_te->pNext )
		{
			num_textures++;
			video_memory_used += p_te->GetVideoMemoryUsage();
			system_memory_used += p_te->GetSystemMemoryUsage();
		}
	}

	*p_num_textures = num_textures;
	*p_video_memory_used = video_memory_used;
	*p_system_memory_used = system_memory_used;
}

//*************************************************************************************
//
//*************************************************************************************
void	ITextureCache::GetOldTextureStats( u32 * p_num_textures,
										   u32 * p_video_memory_used,
										   u32 * p_system_memory_used ) const
{
	u32	num_textures( 0 );
	u32	video_memory_used( 0 );
	u32	system_memory_used( 0 );

	for( TextureEntry * p_te = mpFirstUsedSurface; p_te != NULL; p_te = p_te->pNext )
	{
		num_textures++;
		video_memory_used += p_te->GetVideoMemoryUsage();
		system_memory_used += p_te->GetSystemMemoryUsage();
	}

	*p_num_textures = num_textures;
	*p_video_memory_used = video_memory_used;
	*p_system_memory_used = system_memory_used;
}

//*************************************************************************************
// Purge any textures whos last usage was over 5 seconds ago
//*************************************************************************************
void ITextureCache::PurgeOldTextures()
{
	if (mpTextureHashTable == NULL)
		return;
	
	Lock();

	// XXXX Suspect these need to be a LOT smaller - e.g. potentially 1...
	static const u32 FRAMES_TO_KILL = 5;
	static const u32 FRAMES_TO_DELETE = 30;
	
	for (u32 i = 0; i < mTextureHashSize; i++)
	{
		TextureEntry * p_entry( mpTextureHashTable[i] );
		while (p_entry)
		{
			TextureEntry * p_next( p_entry->pNext );
			
			if ( gRDPFrame - p_entry->FrameLastUsed > FRAMES_TO_KILL )
			{
				RemoveTextureEntry(p_entry);
			}
			p_entry = p_next;
		}
	}
	
	
	// Remove any old textures that haven't been recycled in 1 minute or so
	// Normally these would be reused
	TextureEntry * p_prev( NULL );
	TextureEntry * p_current( mpFirstUsedSurface );
	while( p_current != NULL )
	{
		TextureEntry * p_next( p_current->pNext );
		
		if ( gRDPFrame - p_current->FrameLastUsed > FRAMES_TO_DELETE )
		{
			// Everything from this point on should be too old!
			// Remove from list
			if (p_prev != NULL) p_prev->pNext      = p_current->pNext;
			else				mpFirstUsedSurface = p_current->pNext;
			
			delete p_current;
			
			// p_prev remains the same
			p_current = p_next;	
		}
		else
		{
			p_prev = p_current;
			p_current = p_next;
		}
	}
	
	Unlock();
}

//*************************************************************************************
//
//*************************************************************************************
void ITextureCache::DropTextures()
{
	if (mpTextureHashTable == NULL)
		return;
	
	Lock();
	
	u32 count( 0 );
	u32 total_uses( 0 );
	
	for (u32 i = 0; i < mTextureHashSize; i++)
	{
		while (mpTextureHashTable[i])
		{
			TextureEntry *pTVictim = mpTextureHashTable[i];
			mpTextureHashTable[i] = pTVictim->pNext;
			
			total_uses += pTVictim->NumUses;
			count++;
			
			AddToRecycleList(pTVictim);
		}
	}
	
	DBGConsole_Msg(0, "Texture Handler: %d entries in hash table, %d uses.", count, total_uses);
	
	Unlock();
	
}

//*************************************************************************************
// Add to the recycle list
//*************************************************************************************
void ITextureCache::AddToRecycleList(TextureEntry *p_entry)
{
	if (p_entry->pTexture == NULL)
	{
		// No point in saving!
		delete p_entry;
	}
	else
	{
		// Add to the list
		p_entry->pNext = mpFirstUsedSurface;
		mpFirstUsedSurface = p_entry;
	}
}

//*************************************************************************************
// Search for a texture of the specified dimensions to recycle
//*************************************************************************************
TextureEntry * ITextureCache::ReviveUsedTexture(u32 width, u32 height)
{
	TextureEntry * p_prev( NULL );
	TextureEntry * p_current( mpFirstUsedSurface );
	while (p_current != NULL)
	{
		if (p_current->ti.Width == width &&
			p_current->ti.Height == height)
		{
			// Remove from list
			if (p_prev != NULL) p_prev->pNext      = p_current->pNext;
			else				mpFirstUsedSurface = p_current->pNext;
			
			//DBGConsole_Msg(0, "Reviving used texture (%d x %d)", width, height);
			// Initialise any fields:
			return p_current;
		}
		
		p_prev = p_current;
		p_current = p_current->pNext;
	}

	return NULL;
}

//*************************************************************************************
//
//*************************************************************************************
u32 ITextureCache::GetHashIndex(u32 value)
{
	// Divide by four, because most textures will be on a 4 byte boundry, so bottom four
	// bits are null
	return (value>>2) % mTextureHashSize;
}

//*************************************************************************************
//
//*************************************************************************************
void ITextureCache::AddTextureEntry(TextureEntry *p_entry)
{	
	if (mpTextureHashTable == NULL)
		return;
		
	u32	hash_index( GetHashIndex(p_entry->ti.Address) );
	
	Lock();
	
	// Add to head (not tail, for speed - new textures are more likely to be accessed next)
	p_entry->pNext = mpTextureHashTable[hash_index];
	mpTextureHashTable[hash_index] = p_entry;
	
	Unlock();
	
}

//*************************************************************************************
//
//*************************************************************************************
TextureEntry * ITextureCache::GetEntry(TextureInfo * pti)
{
	if (mpTextureHashTable == NULL)
		return NULL;
	
	// See if it is already in the hash table
	u32 hash_index( GetHashIndex(pti->Address) );
	for( TextureEntry *p_entry = mpTextureHashTable[hash_index]; p_entry != NULL; p_entry = p_entry->pNext )
	{
		if ( p_entry->ti.IsEqual(*pti) )
			return p_entry;
	}
	return NULL;
}

//*************************************************************************************
//
//*************************************************************************************
void ITextureCache::RemoveTextureEntry(TextureEntry * p_entry)
{
	if (mpTextureHashTable == NULL)
		return;
	
	Lock();
	
	//DBGConsole_Msg(0, "Remove Texture entry!");
	
	// See if it is already in the hash table
	u32 hash_index( GetHashIndex(p_entry->ti.Address) );
	
	TextureEntry * p_prev( NULL );
	TextureEntry * p_current( mpTextureHashTable[hash_index] );
	while( p_current != NULL )
	{
		// Check that the attributes match
		if (p_current->ti.IsEqual(p_entry->ti))
		{
			if (p_prev != NULL) p_prev->pNext = p_current->pNext;
			else				mpTextureHashTable[hash_index] = p_current->pNext;

			break;
		}
		
		p_prev = p_current;
		p_current = p_current->pNext;
	}
	
	Unlock();
	
	if (p_current == NULL)
	{
		DBGConsole_Msg(0, "Entry not found!!!");
	}
	
	AddToRecycleList(p_entry);
}

//*************************************************************************************
//
//*************************************************************************************
TextureEntry * ITextureCache::CreateEntry(TextureInfo * pti)
{
	// Find a used texture
	TextureEntry * p_entry( ReviveUsedTexture(pti->Width, pti->Height) );
	if (p_entry == NULL)
	{
		// Couldn't find on - recreate!
		p_entry = new TextureEntry;
		if (p_entry == NULL)
			return NULL;
		
		p_entry->pTexture = CTexture::Create(pti->Width, pti->Height);

		// Ignore failure for now. This prevents us trying to recreate
		// the texture each frame, which would really slow things down
	}
	
	// Initialise
	p_entry->pNext = NULL;
	p_entry->ti = *pti;
	p_entry->NumUses = 0;
	p_entry->FrameLastUsed = gRDPFrame;
	p_entry->HashValue = GenerateHashValue(pti);
	
	// Add to the hash table
	AddTextureEntry(p_entry);
	return p_entry;	
}

//*************************************************************************************
//
//*************************************************************************************
namespace
{
	u32 pixels2bytes( u32 pixels, u32 size )
	{
		return ((pixels << size)+1) / 2;
	}
}

//*************************************************************************************
//
//*************************************************************************************
u32 ITextureCache::GenerateHashValue(TextureInfo * pti) const
{
	// If CRC checking is disabled, always return 0
	if ( !g_bCRCCheck )
		return 0;
	
	u32 bytes_per_line( pixels2bytes( pti->Width, pti->Size ) );

	//DBGConsole_Msg(0, "BytesPerLine: %d", bytes_per_line);
	
	// A very simple crc - just summation
	u32 hash_value( 0 );

	u32 offset( (pti->Top * pti->Pitch) + pixels2bytes( pti->Left, pti->Size ) ); 

	//DAEDALUS_ASSERT( (pti->Address + offset + pti->Height * pti->Pitch) < 4*1024*1024, "Address of texture is out of bounds" );

	const u8 * p_bytes( g_pu8RamBase + pti->Address + offset );
	for (u32 y = 0; y < pti->Height; y++)		// Do every nth line?
	{
		// Byte fiddling won't work, but this probably doesn't matter
		hash_value = daedalus::hash( p_bytes, bytes_per_line, hash_value );
		p_bytes += pti->Pitch;
	}

	if (pti->Format == G_IM_FMT_CI)
	{
		u32 bytes;
		if ( pti->Size == G_IM_SIZ_4b )		bytes = 16  * 4;
		else								bytes = 256 * 4;

		p_bytes = &gTextureMemory[ pti->TmemPalAddress << 3 ];
		hash_value = daedalus::hash( p_bytes, bytes, hash_value );
	}

	return hash_value;
}

//*************************************************************************************
//
//*************************************************************************************
CTexture * ITextureCache::GetTexture(TextureInfo * pti)
{
	DAEDALUS_PROFILE( "ITextureCache::GetTexture" );

	TextureEntry * p_entry( GetEntryAndUpdate(pti) );
	if (p_entry == NULL)
		return NULL;

	return p_entry->pTexture;
}

//*************************************************************************************
// If already in table, return cached copy
// Otherwise, create surfaces, and load texture into memory
//*************************************************************************************
TextureEntry * ITextureCache::GetEntryAndUpdate(TextureInfo * pti)
{
	;
	
	//
	// Retrieve the texture from the cache (if it already exists)
	//
	TextureEntry * p_entry( GetEntry(pti) );
	if (p_entry == NULL)
	{
		// We need to create a new entry, and add it to the hash table.
		p_entry = CreateEntry(pti);
		if (p_entry == NULL)
			return NULL;

	}
	else
	{
		// See if we've already checked that this entry is valid this frame
		if (p_entry->FrameLastUsed == gRDPFrame)
		{
			p_entry->NumUses++;
			return p_entry;
		}

		//
		// Generate and check the current crc
		//
		u32 hash_value( GenerateHashValue(pti) );

		// Check if the CRCs match, if so return. Don't do this for newly created entries
		if (p_entry->HashValue == hash_value)
		{
			// Tile is ok, return
			p_entry->NumUses++;
			p_entry->FrameLastUsed = gRDPFrame;			
			return p_entry;
		}

		// The crc needs updating
		p_entry->HashValue = hash_value;
	}

		
	//
	// The texture needs to be updated for this frame
	//
	if (p_entry->pTexture != NULL)
	{
		DAEDALUS_PROFILE( "Texture Conversion" );

#if RDP_EMULATE_TMEM
		DecompressTmemTexture(p_entry);		
#else
		DecompressTexture(p_entry);		
#endif
	}
	
	return p_entry;
	
}

//*************************************************************************************
// 
//*************************************************************************************
static const char *pszImgFormat[8] = {"RGBA", "YUV", "CI", "IA", "I", "?1", "?2", "?3"};
static const BYTE pnImgSize[4]   = {4, 8, 16, 32};

void ITextureCache::DecompressTexture(TextureEntry * p_entry)
{
	
	BOOL bHandled;	
	
	//void * pSrc = g_pu8RamBase + p_entry->ti.Address;
	
	bHandled = FALSE;
	
	switch (p_entry->ti.Format)
	{
	case G_IM_FMT_RGBA:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_16b:
			ConvertRGBA16(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			ConvertRGBA32(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		}
		break;
		
	case G_IM_FMT_YUV:
		break;

	case G_IM_FMT_CI:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_4b: // 4bpp
			switch (p_entry->ti.TLutFmt)
			{
				//case G_TT_NONE:
			case G_TT_RGBA16:
				ConvertCI4_RGBA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertCI4_IA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
			
		case G_IM_SIZ_8b: // 8bpp
			switch(p_entry->ti.TLutFmt)
			{
			case G_TT_RGBA16:
				ConvertCI8_RGBA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertCI8_IA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
		}
		
		break;

	case G_IM_FMT_IA:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertIA4(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertIA8(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_16b:
			ConvertIA16(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			break;
		}
		break;

	case G_IM_FMT_I:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertI4(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertI8(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		}
		break;
	default:
		break;
	}
	
	if (bHandled)
	{
#ifdef LIBPNG_SUPPORT
		if ( mDumpTextures )
		{
		
			char szFileName[MAX_PATH+1];
			char szFilePath[MAX_PATH+1];
			char szDumpDir[MAX_PATH+1];
			
			wsprintf(szDumpDir, "%s\\%s", TEXT("tile_dumps"), g_ROM.szGameName);
			
			Dump_GetDumpDirectory(szFilePath, szDumpDir);
			
			wsprintf(szFileName, TEXT("%08.8x-%s_%dbpp-%dx%d-%dx%d.png"),
				p_entry->ti.Address, pszImgFormat[p_entry->ti.Format], pnImgSize[p_entry->ti.Size],
				p_entry->ti.Left, p_entry->ti.Top,
				p_entry->ti.Width,p_entry->ti.Height);
			
			IO::Path::Append(szFilePath, szFileName);
			
			p_entry->pTexture->DumpImageAsPNG(szFilePath, p_entry->ti.Width,p_entry->ti.Height);
		}
#endif	
		
	}
	else
	{
		static BOOL bWarningEmitted = FALSE;
		
		if (!bWarningEmitted)
		{			
			DBGConsole_Msg(0, "DecompressTexture: Unable to decompress %s/%dbpp", pszImgFormat[p_entry->ti.Format], pnImgSize[p_entry->ti.Size]);	
			//bWarningEmitted = TRUE;
		}
	}
}

//*************************************************************************************
// Load a texture from tmem
//*************************************************************************************
void ITextureCache::DecompressTmemTexture(TextureEntry * p_entry)
{
	
	BOOL bHandled;	
	
	bHandled = FALSE;
	
	switch (p_entry->ti.Format)
	{
	case G_IM_FMT_RGBA:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_16b:
			ConvertTmemRGBA16(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			ConvertTmemRGBA32(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		}
		break;
		
	case G_IM_FMT_YUV:
		break;

	case G_IM_FMT_CI:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_4b: // 4bpp
			switch (p_entry->ti.TLutFmt)
			{
				//case G_TT_NONE:
			case G_TT_RGBA16:
				ConvertTmemCI4_RGBA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertTmemCI4_IA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
			
		case G_IM_SIZ_8b: // 8bpp
			switch(p_entry->ti.TLutFmt)
			{
			case G_TT_RGBA16:
				ConvertTmemCI8_RGBA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertTmemCI8_IA16(p_entry->pTexture, p_entry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
		}
		
		break;

	case G_IM_FMT_IA:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertTmemIA4(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertTmemIA8(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_16b:
			ConvertTmemIA16(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			break;
		}
		break;

	case G_IM_FMT_I:
		switch (p_entry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertTmemI4(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertTmemI8(p_entry->pTexture, p_entry->ti);
			bHandled = TRUE;
			break;
		}
		break;
	default:
		break;
	}
	
	if (bHandled)
	{
#ifdef LIBPNG_SUPPORT
		if ( mDumpTextures )
		{
			
			char szFileName[MAX_PATH+1];
			char szFilePath[MAX_PATH+1];
			char szDumpDir[MAX_PATH+1];
			
			wsprintf(szDumpDir, "%s\\%s", TEXT("tile_dumps"), g_ROM.szGameName);
			
			Dump_GetDumpDirectory(szFilePath, szDumpDir);
			
			wsprintf(szFileName, TEXT("%08.8x-%s_%dbpp-%dx%d-%dx%d.png"),
				p_entry->ti.Address, pszImgFormat[p_entry->ti.Format], pnImgSize[p_entry->ti.Size],
				p_entry->ti.Left, p_entry->ti.Top,
				p_entry->ti.Width,p_entry->ti.Height);
			
			IO::Path::Append(szFilePath, szFileName);
			
			p_entry->pTexture->DumpImageAsPNG(szFilePath, p_entry->ti.Width,p_entry->ti.Height);
		}
#endif	
		
	}
	else
	{
		static BOOL bWarningEmitted = FALSE;
		
		if (!bWarningEmitted)
		{			
			DBGConsole_Msg(0, "DecompressTexture: Unable to decompress %s/%dbpp", pszImgFormat[p_entry->ti.Format], pnImgSize[p_entry->ti.Size]);	
			//bWarningEmitted = TRUE;
		}
	}
}

//*************************************************************************************
//	Display info on the number of textures in the cache
//*************************************************************************************
void	ITextureCache::DisplayStats()
{
	u32		used_textures;
	u32		used_texture_video_mem;
	u32		used_texture_system_mem;
	u32		old_textures;
	u32		old_texture_video_mem;
	u32		old_texture_system_mem;

	GetUsedTextureStats( &used_textures, &used_texture_video_mem, &used_texture_system_mem );
	GetOldTextureStats( &old_textures, &old_texture_video_mem, &old_texture_system_mem );

	pspDebugScreenPrintf( "Texture Cache Stats\n" );
	pspDebugScreenPrintf( " Used: %3d v:%4dKB s:%4dKB\n", used_textures, used_texture_video_mem / 1024, used_texture_system_mem / 1024 );
	pspDebugScreenPrintf( "  Old: %3d v:%4dKB s:%4dKB\n", old_textures, old_texture_video_mem / 1024, old_texture_system_mem / 1024 );
}
