/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "MainMenuScreen.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "AboutScreen.h"
#include "GlobalSettingsScreen.h"
#include "PspRomSelector.h"
#include "PspRomSettings.h"
#include "ColourPulser.h"

#include "Interface/IniFile.h"

#include "PSPGraphics/FFXMaths.h"
#include "PSPGraphics/DrawText.h"
#include "PSPGraphics/PSPColour.h"
#include "PSPGraphics/GraphicsContext.h"

#include "Core/ROM.h"

#include "DaedMathUtil.h"

#include <pspctrl.h>
#include <pspgu.h>

namespace 
{
	const u32				TEXT_AREA_TOP = 85;
	const u32				TEXT_AREA_LEFT = 40;
	const u32				TEXT_AREA_RIGHT = 480-40;

	const u32				MS_PER_COLOUR_CYCLE = 1200;

	const s32				MAX_TEXT_BRIGHTNESS( 0xee );

	const u32 COLOUR_SELECTION_DIM		= MAKE_COLOUR_U32( 0xe0, 0xe0, 0x80 );
	const u32 COLOUR_SELECTION_BRIGHT	= MAKE_COLOUR_U32( 0xff, 0xff, 0x80 );

	const u32 COLOUR_TEXT_BLUE			= MAKE_COLOUR_U32( 100,100, MAX_TEXT_BRIGHTNESS );
	const u32 COLOUR_TEXT_WHITE			= MAKE_COLOUR_U32( MAX_TEXT_BRIGHTNESS, MAX_TEXT_BRIGHTNESS, MAX_TEXT_BRIGHTNESS );

	const char * const		BACKGROUND_FILENAME = "Resources/background.png";

	enum EMenuOption
	{
		MO_SELECT_ROM = 0,
		MO_SETTINGS,
		MO_ABOUT,
		NUM_MENU_OPTIONS,
	};

	const EMenuOption	MO_FIRST_OPTION = MO_SELECT_ROM;
	const EMenuOption	MO_LAST_OPTION = MO_ABOUT;

	const char * const	gMenuOptionNames[ NUM_MENU_OPTIONS ] =
	{
		"Select Rom",
		"Settings",
		"About",
	};
}

//*************************************************************************************
//
//*************************************************************************************
class IMainMenuScreen : public CMainMenuScreen, public CUIScreen
{
	public:

		IMainMenuScreen();
		~IMainMenuScreen();

		// CPspRomSelector
		virtual void				Run( CUIContext * p_context );

		// CUIScreen
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();
		virtual bool				IsFinished() const									{ return mIsFinished; }

	private:
				s32					GetScrollOffset( u32 item_idx ) const;

	private:
		bool						mIsFinished;
		CColourPulser				mColourPulser;
		s32							mScrollInTime;

		EMenuOption					mCurrentOption;
};

//*************************************************************************************
//
//*************************************************************************************
CMainMenuScreen::~CMainMenuScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
CMainMenuScreen *	CMainMenuScreen::Create()
{
	return new IMainMenuScreen;
}

//*************************************************************************************
//
//*************************************************************************************
IMainMenuScreen::IMainMenuScreen()
:	mIsFinished( false )
,	mColourPulser( COLOUR_SELECTION_DIM, COLOUR_SELECTION_BRIGHT, MS_PER_COLOUR_CYCLE )
,	mScrollInTime( 0 )
,	mCurrentOption( MO_SELECT_ROM )
{
}

//*************************************************************************************
//
//*************************************************************************************
IMainMenuScreen::~IMainMenuScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
void	IMainMenuScreen::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if((new_buttons & PSP_CTRL_START) ||
			(new_buttons & PSP_CTRL_CROSS))
		{
			mIsFinished = true;
		}

		if( new_buttons & PSP_CTRL_UP )
		{
			if( mCurrentOption > MO_FIRST_OPTION )
			{
				mCurrentOption = EMenuOption( mCurrentOption - 1 );
			}
		}
		if( new_buttons & PSP_CTRL_DOWN )
		{
			if( mCurrentOption < MO_LAST_OPTION )
			{
				mCurrentOption = EMenuOption( mCurrentOption + 1 );
			}
		}
	}

	s32		elapsed_ms = s32( 1000.0f * elapsed_time );

	mColourPulser.Update( elapsed_ms );

	mScrollInTime += elapsed_ms;
}

//*************************************************************************************
//
//*************************************************************************************
void	IMainMenuScreen::Render()
{
	s32		x,y;
	const u32	font_height( CDrawText::GetFontHeight() );

	y = TEXT_AREA_TOP;

	for( u32 i = 0; i < NUM_MENU_OPTIONS; ++i )
	{
		EMenuOption		option = EMenuOption( i );

		const char * p_option_text( gMenuOptionNames[ i ] );
		x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, p_option_text, DrawTextUtilities::AT_CENTRE );

		u32	colour;
		if( option == mCurrentOption )
		{
			colour = mColourPulser.GetCurrentColour();
		}
		else
		{
			colour = COLOUR_TEXT_WHITE;
		}

		s32 scroll_y = y + GetScrollOffset( i );

		DrawTextUtilities::DrawShadowText( x, scroll_y, p_option_text, colour );
		y += font_height;		
	}
}

//*************************************************************************************
//
//*************************************************************************************
s32		IMainMenuScreen::GetScrollOffset( u32 item_idx ) const
{
	// At time 0, offset is 
	const s32		TIME_TO_SCROLL_IN = 500;
	const s32		TIME_BETWEEN_ITEMS = 120;
	const s32		OFFSCREEN_OFFSET = 272 - TEXT_AREA_TOP;

	s32				scroll_time( mScrollInTime - (item_idx * TIME_BETWEEN_ITEMS) );
	float			time_fraction( Clamp( float( scroll_time ) / float(TIME_TO_SCROLL_IN), 0.0f, 1.0f ) );
	float			theta( (time_fraction + 1 ) * PI / 2.0f );
	float			scroll_offset( (cosf( theta ) + 1.0f) * OFFSCREEN_OFFSET );

	return s32( scroll_offset );
}

//*************************************************************************************
//
//*************************************************************************************
void	IMainMenuScreen::Run( CUIContext * p_context )
{
	bool	rom_selected( false );
	while( !rom_selected )
	{
		mScrollInTime = 0;
		mIsFinished = false;
		p_context->SetBackgroundTexture( BACKGROUND_FILENAME );
		p_context->Run( this );

		switch( mCurrentOption )
		{
			case MO_SELECT_ROM:
			{
				char		rom_filename[MAX_PATH+1] = "";

				CPspRomSelector *	p_selector( CPspRomSelector::Create() );
				p_selector->Run( p_context );	
				p_selector->GetSelectedRom( rom_filename );
				delete p_selector;

				bool settings_confirmed = false;
				while( !settings_confirmed )
				{
					RomID		rom_id;
					u32			rom_size;
					ECicType	boot_type;

					if(ROM_GetRomID( rom_filename, &rom_id, &rom_size, &boot_type ))
					{
						RomSettings		settings;

						if ( !CIniFile::Get()->GetSettings( rom_id, &settings ) )
						{
							ROM_ResetSettings( &settings );
						}

						CPspRomSettings *	p_settings( CPspRomSettings::Create() );
						p_settings->SetSettings( &settings );
						if( p_settings->Run( p_context ) == CPspRomSettings::SS_OK )
						{
							p_settings->GetSettings( &settings );

							// XXXX Move this somewhere else.

							// Clear everything to black - looks a bit tidier
							for( u32 i = 0; i < 2; ++i )
							{
								CGraphicsContext::Get()->BeginFrame();
								CGraphicsContext::Get()->Clear( true, true );
								CGraphicsContext::Get()->EndFrame();
								CGraphicsContext::Get()->UpdateFrame( false );
							}

							ROM_LoadFile( rom_filename, rom_id, settings );
							rom_selected = true;
						}
						else
						{
							// Back selected - reenter rom select screen
						}

						delete p_settings;
					}
					else
					{
						// XXXX Problem retrieving rom info- should report this!
					}
					
				}
				break;
			}
			case MO_SETTINGS:
			{
				CGlobalSettingsScreen *	p_settings( CGlobalSettingsScreen::Create() );
				p_settings->Run( p_context );	
				delete p_settings;
				break;
			}

			case MO_ABOUT:
			{
				CAboutScreen *	p_about( CAboutScreen::Create() );
				p_about->Run( p_context );	
				delete p_about;
				break;
			}
			default:
			{
				DAEDALUS_ERROR( "Unhandled option" );
				break;
			}

		}
	}
}
