// Profiler.cpp: implementation of the CProfiler class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Profiler.h"

#ifdef DAEDALUS_ENABLE_PROFILING

#include "Debug/DBGConsole.h"
#include "DaedTiming.h"

#include <vector>
#include <string>
#include <stack>
#include <set>

//*************************************************************************************
//
//*************************************************************************************
class CProfileItem
{
	public:
		typedef std::vector< CProfileItem * > ItemVector;

		CProfileItem( const char * p_str )
			:	mName( p_str )
			,	mpParent( NULL )
			,	mTotalTime( 0 )
			,	mStartTime( 0 )
			,	mHitCount( 0 )
		{
		}

		const char *		GetName() const					{ return mName.c_str(); }

		void				Reset();

		void				AddChild( CProfileItem * p_child );
		const ItemVector &	GetChildren() const				{ return mChildren; }
		CProfileItem *		GetParent() const				{ return mpParent; }
		u32					GetDepth() const;

		void				StartTiming();
		void				StopTiming();

		u64					GetTotalTime() const			{ return mTotalTime; }
		u32					GetHitCount() const				{ return mHitCount; }

	private:
		std::string					mName;

		CProfileItem *				mpParent;
		std::vector< CProfileItem * >mChildren;

		u64							mTotalTime;				// The total time accumulated since the last Reset
		u64							mStartTime;				// The time when StartTiming() was called
		u32							mHitCount;				// The total number of times StartTiming() has been called since the last Reset()

};


//*************************************************************************************
//
//*************************************************************************************
class CProfilerImpl
{
	public:
		CProfilerImpl();
		~CProfilerImpl();

		void					Display();
		void					Update();

		SProfileItemHandle		AddItem( const char * p_str );

		void					Enter( SProfileItemHandle handle );
		void					Exit( SProfileItemHandle handle );

	private:
		typedef std::vector< CProfileItem * >			ProfileItemList;
		typedef std::stack< CProfileItem * >			ProfileItemStack;

		f32						mFrequencyInv;
		ProfileItemList			mItems;
		ProfileItemStack		mActiveItems;
};


//*************************************************************************************
//
//*************************************************************************************
CProfilerImpl::CProfilerImpl()
{
	u64	frequency;
	NTiming::GetPreciseFrequency( &frequency );
	mFrequencyInv = 1.0f / float( frequency );
}

//*************************************************************************************
//
//*************************************************************************************
CProfilerImpl::~CProfilerImpl()
{
	for( ProfileItemList::iterator it = mItems.begin(); it != mItems.end(); ++it )
	{
		CProfileItem * p_item( *it );
		delete p_item;
	}
}

//*************************************************************************************
//
//*************************************************************************************
SProfileItemHandle		CProfilerImpl::AddItem( const char * p_str )
{
	// Create a new item
	SProfileItemHandle	handle( mItems.size() );
	CProfileItem *		p_item( new CProfileItem( p_str ) );

	mItems.push_back( p_item );

	return handle;
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfilerImpl::Display()
{
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfilerImpl::Update()
{
	DAEDALUS_ASSERT( mActiveItems.empty(), "There are still active items" );

	while( !mActiveItems.empty() )
	{
		mActiveItems.pop();
	}

	std::stack< CProfileItem * >	item_stack;
	std::set< CProfileItem * >		visited;

	// Initially, push back all the root items
	for( ProfileItemList::const_iterator it = mItems.begin(); it != mItems.end(); ++it )
	{
		CProfileItem * p_item( *it );

		if( p_item->GetParent() == NULL )
		{
			item_stack.push( p_item );
		}
	}

	const char * const TERMINAL_SAVE_CURSOR			= "\033[s";
	const char * const TERMINAL_RESTORE_CURSOR		= "\033[u";
//	const char * const TERMINAL_TOP_LEFT			= "\033[2A\033[2K";
	const char * const TERMINAL_TOP_LEFT			= "\033[H\033[J";

	printf( TERMINAL_SAVE_CURSOR );
	printf( TERMINAL_TOP_LEFT );

	while( !item_stack.empty() )
	{
		CProfileItem *						p_item( item_stack.top() );
		const CProfileItem::ItemVector &	children( p_item->GetChildren() );

		item_stack.pop();
		visited.insert( p_item );

		// Push back all the unvisited children (in reverse order)
		for( CProfileItem::ItemVector::const_reverse_iterator ch_it = children.rbegin(); ch_it != children.rend(); ++ch_it )
		{
			CProfileItem *	p_child( *ch_it );
			if( visited.find( p_child ) == visited.end() )
			{
				item_stack.push( p_child );
			}
		}

		u32		hit_count( p_item->GetHitCount() );

		if( hit_count > 0 )
		{
			// Display details on this item
			u32		depth( p_item->GetDepth() );
			u32		total_ms( u32( p_item->GetTotalTime() * 1000.0f * mFrequencyInv ) );

			printf( "\033[2K%*s %s %dms (%d calls)\n", depth, "", p_item->GetName(), total_ms, hit_count );
			//DBGConsole_Msg( 0, "%*s %s %dms (%d calls)", depth, "", p_item->GetName(), total_ms, hit_count );
		}
	}

	printf( "\033[2K----------------\n\033[2K\n" );
	printf( TERMINAL_RESTORE_CURSOR );
	fflush( stdout );

	for( ProfileItemList::iterator it = mItems.begin(); it != mItems.end(); ++it )
	{
		CProfileItem * p_item( *it );
		p_item->Reset();
	}
}

//*************************************************************************************
// Start profiling for an item
//*************************************************************************************
void	CProfilerImpl::Enter( SProfileItemHandle handle )
{
	DAEDALUS_ASSERT( handle.Handle < mItems.size(), "Invalid handle!" );
	CProfileItem *			p_item( mItems[ handle.Handle ] );

	p_item->StartTiming();

	if( !mActiveItems.empty() )
	{
		CProfileItem * p_parent( mActiveItems.top() );

		p_parent->AddChild( p_item );
	}

	mActiveItems.push( p_item );
}

//*************************************************************************************
// Stop profiling for an item
//*************************************************************************************
void	CProfilerImpl::Exit( SProfileItemHandle handle )
{
	DAEDALUS_ASSERT( handle.Handle < mItems.size(), "Invalid handle!" );
	CProfileItem *			p_item( mItems[ handle.Handle ] );

	p_item->StopTiming();

	if( !mActiveItems.empty() )
	{
		if( p_item == mActiveItems.top() )
		{
			mActiveItems.pop();
		}
		else
		{
			char msg[ 1000 ];
			sprintf( msg, "This item isn't at the top of the stack (expecting %s, found %s)", p_item->GetName(), mActiveItems.top()->GetName() );
			DAEDALUS_ASSERT( p_item == mActiveItems.top(), msg );
	//		DAEDALUS_ASSERT( p_item == mActiveItems.top(), "This item isn't at the top of the stack" );
		}
	}
	else
	{
		DAEDALUS_ERROR( "There is no currently active item!" );
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfileItem::Reset()
{
	mChildren.clear();
	mpParent = NULL;
	mTotalTime = 0;
	mStartTime = 0;
	mHitCount = 0;
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfileItem::AddChild( CProfileItem * p_child )
{
	if( p_child->mpParent == NULL )
	{
		mChildren.push_back( p_child );
		p_child->mpParent = this;
	}
	else
	{
		// We may add the same child more than once per update
		DAEDALUS_ASSERT( p_child->mpParent == this, "Child already has a parent" );
	}
}

//*************************************************************************************
//
//*************************************************************************************
u32	CProfileItem::GetDepth() const
{
	u32				depth( 0 );
	CProfileItem *	p_cur( mpParent );
	while( p_cur != NULL )
	{
		depth++;
		p_cur = p_cur->mpParent;
	}

	return depth;
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfileItem::StartTiming()
{
	DAEDALUS_ASSERT( mStartTime == 0, "Already timing - recursive?" );

	NTiming::GetPreciseTime( &mStartTime );
	mHitCount++;
}

//*************************************************************************************
//
//*************************************************************************************
void CProfileItem::StopTiming()
{
	DAEDALUS_ASSERT( mStartTime != 0, "We're not currently timing" );

	if( mStartTime != 0 )
	{
		u64 end;

		NTiming::GetPreciseTime( &end );

		mTotalTime += end - mStartTime;

		mStartTime = 0;
	}
}

//*************************************************************************************
//
//*************************************************************************************
CProfiler::CProfiler()
:	mpImpl( new CProfilerImpl( ) )
{
}

//*************************************************************************************
//
//*************************************************************************************
CProfiler::~CProfiler()
{
	delete mpImpl;
}

//*************************************************************************************
//
//*************************************************************************************
template<> bool CSingleton< CProfiler >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new CProfiler();

	return true;
}

//*************************************************************************************
//
//*************************************************************************************
SProfileItemHandle		CProfiler::AddItem( const char * p_str )
{
	return mpImpl->AddItem( p_str );
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfiler::Enter( SProfileItemHandle handle )
{
	mpImpl->Enter( handle );
}

//*************************************************************************************
//
//*************************************************************************************
void	CProfiler::Exit( SProfileItemHandle handle )
{
	mpImpl->Exit( handle );
}

//*************************************************************************************
//
//*************************************************************************************
void CProfiler::Update()
{
	mpImpl->Update();
}

//*************************************************************************************
//
//*************************************************************************************
void CProfiler::Display()
{
	mpImpl->Display();
}

#endif
