/*
Copyright (C) 20/02/2004 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "Synchroniser.h"

#include "DaedMathUtil.h"

#include "Interface\MainWindow.h"

#include "Debug\DBGConsole.h"

#include "Core\CPU.h"

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser * CSynchroniser::mpSynchroniser( NULL );

//*****************************************************************************
//
//*****************************************************************************
class ISynchProducer : public CSynchroniser
{
	public:
		ISynchProducer( const CHAR * filename );
		~ISynchProducer();

		virtual ESynchResult	SynchPoint( const void * p_data, u32 length );
		virtual ESynchResult	SynchData( void * p_data, u32 length );

		virtual bool			IsOpen() const;

	private:
				bool			Flush();
		ESynchResult			WriteData( const void * p_data, u32 length );

	private:
		static const u32		BUFFER_SIZE = 4096;
		u8						mBuffer[ BUFFER_SIZE ];
		u32						mBufferCount;
		FILE * const			mFile;
};

class ISynchConsumer : public CSynchroniser
{
	public:
		ISynchConsumer( const CHAR * filename );
		~ISynchConsumer();

		virtual ESynchResult	SynchPoint( const void * p_data, u32 length );
		virtual ESynchResult	SynchData( void * p_data, u32 length );

		virtual bool			IsOpen() const;

	private:
				bool			Fill();

	private:
		static const u32		BUFFER_SIZE = 4096;
		u8						mBuffer[ BUFFER_SIZE ];
		u32						mBufferOffset;
		FILE * const			mFile;
};

//*****************************************************************************
// Creation
//*****************************************************************************
CSynchroniser *	CSynchroniser::CreateProducer( const CHAR * filename )
{
	mpSynchroniser = new ISynchProducer( filename );
	return mpSynchroniser;
}

CSynchroniser *	CSynchroniser::CreateConsumer( const CHAR * filename )
{
	mpSynchroniser = new ISynchConsumer( filename );
	return mpSynchroniser;
}

void CSynchroniser::Destroy()
{
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
void	CSynchroniser::HandleOutOfStorage()
{
	const CHAR *	p_message( "Out of storage for synchroniser" );
	CMainWindow::Get()->MessageBox( p_message );
	DBGConsole_Msg( 0, p_message );
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
void	CSynchroniser::HandleOutOfInput()
{
	const CHAR *	p_message( "Out of input for synchroniser" );
	CMainWindow::Get()->MessageBox( p_message );
	DBGConsole_Msg( 0, p_message );	
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
void	CSynchroniser::HandleOutOfSynch( const CHAR * msg )
{
	const CHAR *	p_message( DSPrintf( "Synchronisation Failed at 0x%08x: %s", gCPUState.CurrentPC, msg ) );
	CMainWindow::Get()->MessageBox( p_message );
	CPU_Halt( p_message );
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
ISynchProducer::ISynchProducer( const CHAR * filename )
:	mFile( fopen( filename, "wb" ) )
,	mBufferCount( 0 )
{
}

//*****************************************************************************
//
//*****************************************************************************
ISynchProducer::~ISynchProducer()
{
	if ( mFile != NULL )
	{
		Flush();
		fclose( mFile );
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool	ISynchProducer::IsOpen() const
{
	return mFile != NULL;
}

//*****************************************************************************
//
//*****************************************************************************
bool	ISynchProducer::Flush()
{
	DAEDALUS_ASSERT( mBufferCount <= BUFFER_SIZE, "How come the buffer has overflowed?" );

	const u32	count( mBufferCount );

	mBufferCount = 0;

	return fwrite( mBuffer, 1, count, mFile ) == count;
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchProducer::WriteData( const void * p_data, u32 length )
{
	if ( mFile != NULL )
	{
		const u8 *	p_current_ptr( reinterpret_cast< const u8 * >( p_data ) );
		u32			bytes_remaining( length );
		while( bytes_remaining > 0 )
		{
			u32		buffer_bytes_remaining( BUFFER_SIZE - mBufferCount );
			u32		bytes_to_process( daedalus::Min( bytes_remaining, buffer_bytes_remaining ) );

			//
			// Append as many bytes as possible
			//
			if( bytes_to_process > 0 )
			{
				memcpy( mBuffer + mBufferCount, p_current_ptr, bytes_to_process );

				p_current_ptr += bytes_to_process;
				bytes_remaining -= bytes_to_process;
				mBufferCount += bytes_to_process;
			}

			//
			// Flush the buffer if full
			//
			if( mBufferCount >= BUFFER_SIZE )
			{
				if( !Flush() )
				{
					return SR_OUT_OF_STORAGE;
				}
			}
		}
		return SR_OK;
	}

	return SR_OUT_OF_STORAGE;
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchProducer::SynchPoint( const void * p_data, u32 length )
{
	return WriteData( p_data, length );
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchProducer::SynchData( void * p_data, u32 length )
{
	return WriteData( p_data, length );
}

//*****************************************************************************
//
//*****************************************************************************
ISynchConsumer::ISynchConsumer( const CHAR * filename )
:	mFile( fopen( filename, "rb" ) )
,	mBufferOffset( BUFFER_SIZE )		// 
{
}

//*****************************************************************************
//
//*****************************************************************************
ISynchConsumer::~ISynchConsumer()
{
	if ( mFile != NULL )
	{
		fclose( mFile );
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool	ISynchConsumer::IsOpen() const
{
	return mFile != NULL;
}

//*****************************************************************************
//
//*****************************************************************************
bool	ISynchConsumer::Fill()
{
	DAEDALUS_ASSERT( mBufferOffset == BUFFER_SIZE, "How come the buffer has overflowed?" );

	mBufferOffset = 0;

	return fread( mBuffer, 1, BUFFER_SIZE, mFile ) == BUFFER_SIZE;
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchConsumer::SynchPoint( const void * p_data, u32 length )
{
	if ( mFile != NULL )
	{
		const u8 *	p_current_ptr( reinterpret_cast< const u8 * >( p_data ) );
		u32			bytes_remaining( length );

		while( bytes_remaining > 0 )
		{
			u32		buffer_bytes_remaining( BUFFER_SIZE - mBufferOffset );
			u32		bytes_to_process( daedalus::Min( bytes_remaining, buffer_bytes_remaining ) );

			if( bytes_to_process > 0 )
			{
				if( memcmp( p_current_ptr, mBuffer + mBufferOffset, bytes_to_process ) != 0 )
				{
					return SR_OUT_OF_SYNCH;
				}

				p_current_ptr += bytes_to_process;
				bytes_remaining -= bytes_to_process;
				mBufferOffset += bytes_to_process;
			}

			if( mBufferOffset >= BUFFER_SIZE )
			{
				if( !Fill() )
				{
					return SR_OUT_OF_INPUT;
				}
			}
		}

		return SR_OK;
	}

	return SR_OUT_OF_INPUT;
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchConsumer::SynchData( void * p_data, u32 length )
{
	if ( mFile != NULL )
	{
		u8 *		p_current_ptr( reinterpret_cast< u8 * >( p_data ) );
		u32			bytes_remaining( length );

		while( bytes_remaining > 0 )
		{
			u32		buffer_bytes_remaining( BUFFER_SIZE - mBufferOffset );
			u32		bytes_to_process( daedalus::Min( bytes_remaining, buffer_bytes_remaining ) );

			if( bytes_to_process > 0 )
			{
				memcpy( p_current_ptr, mBuffer + mBufferOffset, bytes_to_process );

				p_current_ptr += bytes_to_process;
				bytes_remaining -= bytes_to_process;
				mBufferOffset += bytes_to_process;
			}

			if( mBufferOffset >= BUFFER_SIZE )
			{
				if( !Fill() )
				{
					return SR_OUT_OF_INPUT;
				}
			}
		}

		return SR_OK;
	}

	return SR_OUT_OF_INPUT;
}




